#!/usr/bin/env python
# -*- coding:utf-8 -*-
# Author: wxnacy@gmail.com
"""
run 命令的参数解析
"""
import os
from csarg import Action
from csarg import CommandArgumentParserFactory
from csarg.parser import Argument
from wpy.functools import clock

from wush.common import utils
from wush.common.constants import Constants
from wush.common.utils import run_shell
from wush.common.loggers import create_logger
from wush.common.run_mode import RUN_MODE
from wush.common.config_value import environ_keys
from wush.config import load_config
from wush.config.models import RequestModel
from wush.web.cookie import Cookie
from wush.web.enums import RequestsParamsEnum
from wush.web.request import RequestClient
from wush.web.request import RequestBuilder
from wush.web.response import ResponseHandler
from wush.web.history import History

from .command import CmdArgumentParser


@CommandArgumentParserFactory.register()
class RunArgumentParser(CmdArgumentParser):
    cmd = 'run'
    logger = create_logger('RunArgumentParser')

    @classmethod
    def default(cls):
        """
        初始化一个实例
        """
        item = cls()
        item.add_argument('cmd')
        item.add_argument('-m', '--module', help='模块名称')
        item.add_argument('-s', '--space', help='空间名称')
        item.add_argument('-n', '--name', help='请求名称')
        item.add_argument('-c', '--config', help='配置地址')
        item.add_argument('--attr', help='属性信息，自定义')
        item.add_argument('--params', action = Action.APPEND.value,
            help='请求地址参数')
        item.add_argument('--json', action = Action.APPEND.value,
            help='请求 body 参数，json 格式')
        item.add_argument('--env', action = Action.APPEND.value,
            help='传递环境变量')
        item.add_argument('--open', action = Action.STORE_TRUE.value,
            help = '是否通过浏览器打开请求结果')
        item.add_argument('--curl', action = Action.STORE_TRUE.value,
            help = '是否使用 curl 文本')
        item.add_argument('--with-browser-cookie',
            action = Action.STORE_TRUE.value,
            help = '是否使用 curl 文本')
        item.add_argument('--no-handle-response',
            action = Action.STORE_TRUE.value,
            help = '不对 response 做处理')
        item.add_argument('--no-browser',
            action = Action.STORE_TRUE.value,
            help = '不使用浏览器打开结果页面')
        if RUN_MODE.is_command:
            item.add_argument('--url', help='请求地址')

        #  item.config = load_config()

        return item

    def get_completions_after_argument(self, word_for_completion):
        """
        获取补全的单词列表
        :param word_for_completion: 补全需要的单词
        """
        self.config = load_config()
        words = []
        if not self.argument:
            return words
        arg = self.argument
        if word_for_completion == '--name':
            # 针对 --name 参数的自动补全
            requests = self.config.get_requests(arg.module)
            words = []
            for req in requests:
                words.append(dict( text = req.name, display_meta=req.title))
            return words
        elif word_for_completion == '--module':
            modules = self.config.get_modules()
            words = [ dict(text = o.name) for o in modules ]
            return words
        elif word_for_completion == '--params':
            # 参数补全
            request = self.config.get_request(arg.module, arg.name)
            params = request.params.to_dict()
            words = self._dict_to_completions(params)
            return words
        elif word_for_completion == '--json':
            # 参数补全
            request = self.config.get_request(arg.module, arg.name)
            words = self._dict_to_completions(request.json.to_dict())
            return words
        elif word_for_completion == '--env':
            # 环境变量
            # 使用当前请求的环境变量 keys 做补全
            request = self.config.get_request(arg.module, arg.name,
                    set_env=False)
            keys = environ_keys(request)
            self.logger.info(request.path)
            self.logger.info(
                f'request {arg.module} {arg.name} environ_keys {keys}')
            data = { f"{o}": "" for o in keys }

            words = self._dict_to_completions(data)
            return words

        return super().get_completions_after_argument(word_for_completion)

    def _dict_to_completions(self, data):
        """字典转为自动补全信息"""
        words = []
        for k, v in data.items():
            words.append(dict(
                text = '{}='.format(k),
                display = '{}={}'.format(k, v),
                display_meta=''))
        return words

    def run_command(self, text):
        """运行命令行模式"""
        self.config = load_config()
        args = self.parse_args(text)
        # curl 模式下打开一个文件并输入文本供后续使用
        builder: RequestBuilder = None
        if args.curl:
            builder = RequestBuilder()
            filepath = Constants.build_tmpfile('curl')
            run_shell(f'echo "# 请输入 cUrl 文本\n" > {filepath}')
            vim_cmd = f'vim {filepath}'
            os.system(vim_cmd)
            builder = RequestBuilder.load_curl(filepath)

        if args.url:
            builder = RequestBuilder(url = args.url)

        if args.name:
            builder = self._get_request_builder(args)
        builder.argument = args
        builder.set_run_mode(RUN_MODE.mode)

        #  self.logger.info('builder {}'.format(builder.json()))

        req_client = RequestClient(builder)
        res = req_client.request()
        self._run(args, builder, res)

    def _get_request_builder(self, args: Argument) -> RequestBuilder:
        """获取请求构造体"""
        params = utils.list_key_val_to_dict(args.params or [])
        json_data = utils.list_key_val_to_dict(args.json or [])
        environs = utils.list_key_val_to_dict(args.env or [])
        self.logger.info(f'arg params {params}')
        self.logger.info(f'arg json {json_data}')
        self.logger.info(f'arg env {environs}')

        request_model = self.config.get_request(args.module, args.name,
                environs = environs)
        request_model.add_params(**params)
        request_model.add_json(**json_data)

        builder = self._load_builder_from_request(request_model,
            args.with_browser_cookie)
        builder.argument = args
        return builder

    def run_shell(self, text):
        self.config = load_config()
        args = self.parse_args(text)

        if not args.name:
            raise Exception

        builder = self._get_request_builder(args)
        builder.set_run_mode(RUN_MODE.mode)
        request_client = RequestClient(builder)

        self._print('Space: {}'.format(args.space))
        self._print('Module: {}'.format(args.module))
        self._print('Request: {}'.format(args.name))
        self._print('Url: {}'.format(builder.url))
        self._print('Params: {}'.format(builder.params))
        self._print('Json: {}'.format(builder.json))
        self._print('请求中。。。')

        res = request_client.request()

        self._print('Status: {}'.format(res.status_code))

        self._run(args, builder, res)

    def _run(self, args, builder, res):

        # 保存历史记录
        History().save(res)

        if args.open:
            self._print('See in browser')
            self._open(builder.version)
            return

        handler_func = ResponseHandler.get_handler(args.module,
            args.name)
        if handler_func and not args.no_handle_response:
            handler_func(res)
            return

        self.config.function.handler_response( res)

    def _open(self, version):
        #  """打开请求信息"""
        port = self.config.server_port
        url = f"http://localhost:{port}/api/version/{version}"
        utils.open_url(url)

    @classmethod
    @clock(fmt = Constants.CLOCK_FMT, logger_func = logger.info)
    def _load_builder_from_request(cls, request_model: RequestModel,
            with_browser_cookie: bool = False
        ) -> RequestBuilder:
        """加载 RequestModel 模型"""
        request_dict = request_model.dict()
        ins_dict = {}
        for key in RequestsParamsEnum.values():
            val = request_dict.get(key)
            ins_dict[key] = val

        ins = RequestBuilder(**ins_dict)
        request_cookies = {}
        # 对 cookie_domains 进行解析
        cookie_domains = request_model.cookie_domains
        if cookie_domains and with_browser_cookie:
            cookies  = Cookie.get_browser_cookie(*cookie_domains)
            request_cookies.update(cookies)
        request_cookies.update(ins.cookies)
        ins.cookies = request_cookies

        return ins



#!/usr/bin/env python
# -*- coding:utf-8 -*-
# Author: wxnacy@gmail.com
"""
run 命令的参数解析
"""

from .command import CmdArgumentParser

class DefaultArgumentParser(CmdArgumentParser):
    cmd = 'default'

    @classmethod
    def default(cls):
        """
        初始化一个实例
        """
        item = cls()
        item.add_argument('cmd')
        item.add_argument('--config')
        item.add_argument('--module')
        item.add_argument('--space')
        return item




#!/usr/bin/env python
# -*- coding:utf-8 -*-
# Author: wxnacy@gmail.com
"""
工具模块
"""

import json
import os
import importlib
from importlib.machinery import SourceFileLoader
import subprocess

from wush.common.exceptions import JsonException

__all__ = ['get_current_module_path']

def filter_json(data, rules):
    """
    过滤 json 数据
    :param dict data: 过滤的数据
    :param list rules: 过滤规则，例：{"name"}

    最终拼成的命令格式如下
    echo '{"id": 1, "name": "wxnacy"}' | jq '{"name"}'
    """
    cmd_fmt = "echo '{}' | jq '{}'"
    text = json.dumps(data)
    for rule in rules:
        cmd = cmd_fmt.format(text , rule)
        content, err = run_shell(cmd)
        if err:
            raise JsonException(err)
        return json.loads(content)
    return data

def fmt_path(path):
    """格式化地址"""
    path = os.path.expanduser(path)
    return path

def list_key_val_to_dict(data):
    """
    键值对列表转为字典
    :param data: ['name=wxnacy', 'key=True', 'test']
    :returns { 'name': 'wxnacy', 'key': 'True'  }
    """
    res = {}
    for item in data:
        if '=' in item:
            k, v = item.split('=', 1)
            res[k] = v
    return res

def run_shell(command):
    """运行 shell 语句"""
    res = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE,
        stderr = subprocess.PIPE)
    return res.communicate()

def load_module(module_name):
    """加载模块"""
    if module_name.endswith('.py'):
        name = str(hash(module_name))
        views_module = SourceFileLoader(name,
            module_name).load_module()
    else:
        views_module = importlib.import_module(module_name)
    return views_module

def get_current_module_path():
    """获取当前模块的路径"""
    import wush as _wush
    module_path = _wush.__path__[0]
    return module_path

def open_url(url):
    """打开 url"""
    # TODO 适配更多系统
    os.system(f'open -a "/Applications/Google Chrome.app" "{url}"')



#!/usr/bin/env python
# -*- coding:utf-8 -*-
# Author: wxnacy@gmail.com
"""

"""
import sys
import os
#  import re

from csarg import CommandArgumentParserFactory
#  from wpy.path import walkfile
from wush.common.utils import get_current_module_path
from wush.common.utils import load_module

from wush.argument.command import CmdArgumentParser
from wush.common.constants import Constants
from wush.common.loggers import create_logger
from wush.common.run_mode import RUN_MODE
from wush.config import load_config

__all__ = ['Command']

class Command(object):
    logger = create_logger('Command')

    def __init__(self):
        # 初始化创建临时文件目录
        if not os.path.exists(Constants.TMPDIR):
            os.makedirs(Constants.TMPDIR)

        # 动态加载参数解析器模板
        argument_dir = os.path.join(get_current_module_path(), 'argument')
        for name in os.listdir(argument_dir):
            if not name.endswith('.py'):
                continue
            module = f'wush.argument.{name[:-3]}'
            load_module(module)

    def convert_argparse(self, cmd):
        """转换参数解析器"""
        for clz in CmdArgumentParser.__subclasses__():
            if clz.cmd == cmd:
                return clz.default()

        return CommandArgumentParserFactory.build_parser(cmd)

    #  @profile
    def run(self):
        sys_args = sys.argv[1:]
        if not sys_args:
            sys_args = ['shell']
        cmd = sys_args[0]

        if cmd not in CommandArgumentParserFactory.get_cmd_names():
            # 不在程序范围内的命令，默认当做 url 处理
            if not cmd.startswith('http'):
                sys_args[0] = 'http://' + cmd
            sys_args.insert(0, 'run')
            sys_args.insert(1, '--url')
            cmd = 'run'
            #  args_text = ' '.join(sys_args)
            #  # 判断 cmd 是否为 url 格式
            #  if re.match(r'^https?:/{2}\w.+$', cmd):
                #  args_text = 'run ' + args_text.replace(cmd, f'--url {cmd}')
                #  cmd = 'run'

        RUN_MODE.set_command()
        # 转换参数解析器
        parser = self.convert_argparse(cmd)

        args = parser.parse_args(sys_args)
        # 加载配置
        config_path = None
        if hasattr(args, 'config'):
            config_path = args.config
        self.config = load_config(config_path)
        if hasattr(args, 'module'):
            self.config.module_name = args.module
        if hasattr(args, 'space'):
            self.config.space_name = args.space

        parser.run(sys_args)




#!/usr/bin/env python
# -*- coding:utf-8 -*-
# Author: wxnacy@gmail.com
"""

"""

from wush.web.request import RequestClient
from wush.web.request import RequestBuilder
import re


_REG_ENV = r'\{(.+?)\}'


def replace(match):
    print(dir(match))
    print(match.groups())


    print('-' * 10)
    return 'ssss'

def main():
    #  builder = RequestBuilder(method='get', url = 'https://ipconfig.io/json')
    #  client = RequestClient(builder)
    #  res = client.request()
    #  print(res.json())

    text = 'my name is ${name} and age is {age()}'
    text = 'ss'
    print('s')

    print(re.findall(_REG_ENV, text))
    res = re.sub(_REG_ENV, replace, text)
    print(res)
    #  logger.info(res)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding:utf-8 -*-
# Author: wxnacy@gmail.com
"""
run 命令的参数解析
"""
import os
import json
from csarg import Action
from csarg import CommandArgumentParserFactory

from wush.common import utils
from wush.common.config_value import ConfigValue
from wush.common.loggers import create_logger

from .command import CmdArgumentParser


@CommandArgumentParserFactory.register()
class AliasArgumentParser(CmdArgumentParser):
    cmd = 'alias'
    logger = create_logger('AliasArgumentParser')

    @classmethod
    def default(cls):
        """
        初始化一个实例
        """
        item = cls()
        item.add_argument('cmd')
        item.add_argument('-v', '--value', help='命令参数')
        item.add_argument('-n', '--name', help='命令别名')
        item.add_argument('--env', action = Action.APPEND.value,
            help='传递环境变量')

        item.alias_path = os.path.expanduser('~/.wush_aliases')

        return item

    def get_completions_after_argument(self, word_for_completion):
        """
        获取补全的单词列表
        :param word_for_completion: 补全需要的单词
        """
        words = []
        if not self.argument:
            return words
        arg = self.argument
        if word_for_completion == '--name':
            # TODO 针对 --name 参数的自动补全
            requests = self.config.get_requests(arg.module)
            words = []
            for req in requests:
                words.append(dict( text = req.name, display_meta=req.title))
            return words
        return super().get_completions_after_argument(word_for_completion)

    def run_command(self, args):
        """运行命令行模式"""
        args = self.parse_args(args)

        data = self.get_aliases()
        if args.value:
            if not args.name:
                print("缺少必要参数 --name")
                return
            data[args.name] = args.value
            text = '\n'.join([json.dumps({k: v}) for k, v in data.items()])
            with open(self.alias_path, 'w') as f:
                f.write(text)
            return

        if args.name:
            environs = utils.list_key_val_to_dict(args.env or [])
            value = data.get(args.name)
            if not value:
                print(f"alias: {args.name} not found")
                return
            value = ConfigValue(value).set_env(**environs).format()
            os.system(f'wush {value}')
            return

        for name, value in data.items():
            print(f'{name}={value}')

    def get_aliases(self):
        if not os.path.exists(self.alias_path):
            return {}
        with open(self.alias_path, 'r') as f:
            lines = f.readlines()

        data = {}
        for line in lines:
            try:
                item = json.loads(line)
                data.update(item)
            except Exception as e:
                print(e)

        return data

    def run_shell(self, args):
        self.run_command(args)




#!/usr/bin/env python
# -*- coding:utf-8 -*-
# Author: wxnacy(wxnacy@gmail.com)
"""
网络相关工具方法
"""

import argparse
from shlex import split
from http.cookies import SimpleCookie
from urllib.parse import urlparse
from w3lib.http import basic_auth_header

from wush.common.loggers import get_logger

logger = get_logger('curl')

class CurlParser(argparse.ArgumentParser):
    def error(self, message):
        error_msg = \
            'There was an error parsing the curl command: {}'.format(message)
        raise ValueError(error_msg)

curl_parser = CurlParser()
curl_parser.add_argument('url')
curl_parser.add_argument('-H', '--header', dest='headers', action='append')
curl_parser.add_argument('-X', '--request', dest='method', default='get')
curl_parser.add_argument('-d', '--data', '--data-raw', dest='data')
curl_parser.add_argument('-u', '--user', dest='auth')

safe_to_ignore_arguments = [
    ['--compressed'],
    # `--compressed` argument is not safe to ignore, but it's included here
    # because the `HttpCompressionMiddleware` is enabled by default
    ['-s', '--silent'],
    ['-v', '--verbose'],
    ['-#', '--progress-bar']
]
for argument in safe_to_ignore_arguments:
    curl_parser.add_argument(*argument, action='store_true')

def curl_to_request_kwargs(curl_command, ignore_unknown_options=True):
    """Convert a cURL command syntax to Request kwargs.
    :param str curl_command: string containing the curl command
    :param bool ignore_unknown_options: If true, only a warning is emitted when
    cURL options are unknown. Otherwise raises an error. (default: True)
    :return: dictionary of Request kwargs
    """
    curl_args = split(curl_command)
    if curl_args[0] != 'curl':
        raise ValueError('A curl command must start with "curl"')
    parsed_args, argv = curl_parser.parse_known_args(curl_args[1:])
    if argv:
        msg = 'Unrecognized options: {}'.format(', '.join(argv))
        if ignore_unknown_options:
            #  warnings.warn(msg)
            logger.warn(msg)
        else:
            raise ValueError(msg)
    url = parsed_args.url
    # curl automatically prepends 'http' if the scheme is missing, but Request
    # needs the scheme to work
    parsed_url = urlparse(url)
    if not parsed_url.scheme:
        url = 'http://' + url
    result = {'method': parsed_args.method.upper(), 'url': url}
    headers = {}
    cookies = {}
    for header in parsed_args.headers or ():
        name, val = header.split(':', 1)
        name = name.strip()
        val = val.strip()
        if name.title() == 'Cookie':
            for name, morsel in SimpleCookie(val).items():
                cookies[name] = morsel.value
        else:
            headers[name] = val
    if parsed_args.auth:
        user, password = parsed_args.auth.split(':', 1)
        headers['Authorization']= basic_auth_header(user, password)
    if headers:
        result['headers'] = headers
    if cookies:
        result['cookies'] = cookies
    if parsed_args.data:
        result['body'] = parsed_args.data
    return result

class cUrl(object):

    @classmethod
    def dump(cls, filepath):
        with open(filepath, 'r') as f:
            lines = f.readlines()
        texts = []
        for line in lines:
            line = line.strip().strip('\n').strip('\\')
            if line.startswith('#'):
                continue
            texts.append(line)
        text = ''.join(texts)
        return cls.dumps(text)

    @classmethod
    def dumps(cls, text):
        return curl_to_request_kwargs(text)



#!/usr/bin/env python
# -*- coding:utf-8 -*-
# Author: wxnacy@gmail.com
"""
命令补全
"""

from prompt_toolkit.completion import Completion

from csarg import CommandArgumentParserFactory
from .base import BaseCompleter
from .word import WordCompleter
from .filesystem import ExecutableCompleter

from wush.common.loggers import create_logger
from wush.config import load_config

__all__ = ['CommandCompleter']

class CommandCompleter(BaseCompleter):
    """shell 环境补全管理"""
    logger = create_logger("CommandCompleter")

    def __init__(self, argparser):
        self.argparser = argparser
        self.path_completer = ExecutableCompleter()
        self.config = load_config()

    def yield_words(self, words):
        """获取命令参数的补全器"""
        if words and isinstance(words[0], dict):
            words = [Completion(**o) for o in words]
        _completer = WordCompleter(words)
        yield from self.yield_completer(_completer)

    def get_completions(self, document, complete_event):
        super().get_completions(document, complete_event)
        self.document = document
        self.complete_event = complete_event
        self.logger.info(f'document.text {document.text}')
        try:
            # 没输入命令时，将命令列表作为补全列表
            if not document.text:
                all_cmds = list(CommandArgumentParserFactory.get_cmd_names())
                # 补全命令
                yield from self.yield_words(all_cmds)
                return

            self.argparser = CommandArgumentParserFactory.build_parser(document.text)
            self.logger.info('completer argparser %s', self.argparser.cmd)
            arg = self.argparser.parse_args(document.text)
            self.logger.info('args %s', arg)
            cmd = self.first_word

            all_cmds = list(CommandArgumentParserFactory.get_cmd_names())
            # 补全命令
            if cmd not in all_cmds:
                yield from self.yield_words(all_cmds)
                return

            word_for_completion = self.word_for_completion

            # 使用自定义方法返回补全单词
            words = self.config.function.get_completion_words(
                word_for_completion)
            if words:
                yield from self.yield_completer(WordCompleter(words))
                return

            # 补全参数后的信息
            words = self.argparser.get_completions_after_argument(
                    word_for_completion)
            if words:
                yield from self.yield_words(words)
                return

            # 使用模块自带的补全
            if word_for_completion in ('--config', '--root'):
                yield from self.yield_completer(self.path_completer)
            else:
                words = self.argparser.get_completions_after_cmd(arg,
                    word_for_completion)
                yield from self.yield_words(words)

        except:
            import traceback
            self.logger.error(traceback.format_exc())
            self.logger.error(traceback.format_stack())



#!/usr/bin/env python
# -*- coding:utf-8 -*-
# Author: wxnacy(wxnacy@gmail.com)
"""
基础模块
"""

from wpy.base import BaseObject
from wpy.base import BaseEnum

# TODO 改为抽象类
class DataType(BaseObject):
    _value = None       # 对象实例化或手动设置的值
    _type = object      # 字段赋值时应该给予的类型
    _default = None     # 当前类型的默认值
    _name = None        # 字段名称

    default = None      # 使用时赋值的默认数据
    convert = False     # 是否强制转换当前类型

    @property
    def name(self):
        return self._name

    def set_value(self, value):
        self._value = value
        # 强制转换类型
        if self.convert:
            self._value = self._type(self._value)

    def value(self):
        """获取数据"""
        val = str(self.default()) if callable(self.default) else self.default
        # 只有等 None 的时候才会返回默认值
        if self._value == None:
            return val or self._default
        return self._value # or val or self._default

    def clear(self):
        """清空数据"""
        self._value = None

    @property
    def _value_fmt(self):
        return f'{self.name}:{self._value}'

    def valid(self):
        """校验赋值"""
        if self._value != None and not isinstance(self._value, self._type):
            raise ValueError(f'{self._value_fmt} is not be {self._type}')

    @classmethod
    def get_base_datatype(cls, basetype):
        """获取基础类型类"""
        for clz in cls.__subclasses__():
            if clz._type == basetype:
                return clz
        raise ValueError(f'{basetype} is not basetype')


class Str(DataType):
    _type = str
    _default = str()

    enum = None
    upper = False

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if self.enum and not issubclass(self.enum, BaseEnum):
            raise ValueError('enum must be wpy.base.BaseEnum')

    def set_value(self, value):
        super().set_value(value)
        if self.upper and self._value:
            self._value = self._value.upper()

    def valid(self):
        super().valid()
        if self.enum and self._value != None and \
                self._value not in self.enum.values():
            raise ValueError(f'{self._value_fmt} is not {self.enum}')


class Int(DataType):
    _type = int
    _default = int()


class Float(DataType):
    _type = float
    _default = float()


class Dict(DataType):
    _type = dict
    _default = dict()


class List(DataType):
    _type = list
    _default = list()

    model = None

    def value(self):
        """获取数据"""
        if self.model and self._value:
            items = []
            for item in self._value:
                items.append(self.model(**item))
            return items

        return super().value()

class Object(DataType):
    _type = object

    model = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if self.model:
            self._default = self.model()

    def valid(self):
        """校验"""
        super().valid()
        if self._value != None and self.model:
            # 如果当前类型已经格式化，不在执行检查
            if isinstance(self._value, self.model):
                return
            if not isinstance(self._value, dict):
                raise ValueError(f'{self._value_fmt} can not to {self.model}')

    def value(self):
        """获取数据"""
        if self.model and self._value != None:
            # 对于已经格式化的数据不再重复执行
            if isinstance(self._value, self.model):
                return self._value
            return self.model(**self._value)

        return super().value()



#!/usr/bin/env python
# -*- coding:utf-8 -*-
# Author: wxnacy@gmail.com
"""
常量模块
"""

import os
import random

from .utils import get_current_module_path

__all__ = ['Constants']

INIT_CONIFG_YML = """
env:
  github_domain: 'api.github.com'
modules:
  - name: github_user
    domain: '${github_domain}'
    protocol: https
    url_prefix: /users
    requests:
      - name: user_profile
        path: /psf
"""

INIT_CONIFG_TEXT = f"""
# wush 工具的配置文档
# env:
{INIT_CONIFG_YML}
"""

class Constants(object):
    TMPDIR = '/tmp/wush'
    BASETYPE_MAP = {
        'str': str,
        'string': str,
        'int': int,
        'list': list,
        'dict': dict,
        'float': float,
    }
    CONFIG_DIR = os.path.expanduser('~/.wush')
    CONFIG_PATH = os.path.join(CONFIG_DIR, 'config.yml')
    HISTORY_PATH = os.path.expanduser('~/.wush_history')
    API_HISTORY_DIR = os.path.expanduser('~/.wush_api_history')
    SERVER_HOST = 'localhost'
    SERVER_PORT = '6060'
    INIT_CONIFG_YML = INIT_CONIFG_YML
    INIT_CONIFG_TEXT = INIT_CONIFG_TEXT
    CLOCK_FMT = '[耗时][{T:0.4f}s] {F}({A}, {K})'

    @classmethod
    def get_sys_config_path(cls):
        return os.path.join(get_current_module_path(), 'config/config.yml')

    @classmethod
    def build_tmpfile(cls, prefix):
        """创建临时文件"""
        return os.path.join(cls.TMPDIR, f'{prefix}.{random.randint(1, 90000)}')

    @classmethod
    def str_to_basetype(cls, text):
        """字符串转为基础类型"""
        if text not in cls.BASETYPE_MAP:
            raise ValueError(f'{text} is not basetype')
        return cls.BASETYPE_MAP[text]




#!/usr/bin/env python
# -*- coding:utf-8 -*-
# Author: wxnacy(wxnacy@gmail.com)
"""
网络相关工具方法
"""

import socket
import traceback
from wush.common.loggers import get_logger

logger = get_logger()


def telnet(server_ip: str, port: int):
    """链接服务"""
    sk = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sk.settimeout(1) #设置超时时间
    try:
        sk.connect((server_ip,port))
        sk.close()
        return True
    except Exception:
        logger.error(traceback.format_exc())
        logger.error(traceback.format_stack())
        sk.close()
        return False



#!/usr/bin/env python
# -*- coding:utf-8 -*-
# Author: wxnacy(wxnacy@gmail.com)
"""
网络相关枚举模块
"""

from wpy.base import BaseEnum

class MethodEnum(BaseEnum):
    GET = 'GET'
    POST = 'POST'
    PUT = 'PUT'
    DELETE = 'DELETE'
    HEAD = 'HEAD'
    OPTIONS = 'OPTIONS'
    PATCH = 'PATCH'

class ProtocolEnum(BaseEnum):
    HTTP = 'http'
    HTTPS = 'https'

class HeaderEnum(BaseEnum):
    CONTENT_TYPE = 'Content-Type'
    CONTENT_LENGTH = 'Content-Length'
    LOCATION = 'Location'

class ContentTypeEnum(BaseEnum):
    APPLICATION_JSON = 'application/json'
    TEXT_HTML = 'text/html'

class RequestsParamsEnum(BaseEnum):
    METHOD = 'method'
    URL = 'url'
    PARAMS = 'params'
    JSON = 'json'
    HEADERS = 'headers'
    COOKIES = 'cookies'



#!/usr/bin/env python
# -*- coding:utf-8 -*-
# Author: wxnacy(wxnacy@gmail.com)
"""
配置模型
"""
import os
import json
from collections import defaultdict
from typing import (
    Dict, Any, List, Union
)
from pydantic.dataclasses import dataclass
from pydantic import (
    BaseModel as PydanticModel, Field, validator, root_validator
)

from wpy.base import BaseObject

from wush.common.config_value import ConfigValue
from wush.common.constants import Constants
from wush.common.loggers import get_logger
from wush.web.enums import MethodEnum
from wush.web.enums import ProtocolEnum

logger = get_logger('config.models')

class PydanticConfig:
    arbitrary_types_allowed = True

class BaseModel(BaseObject):

    def inherit(self, super_ins, keys):
        """继承数据"""
        for key in keys:
            super_value = getattr(super_ins, key)
            sub_value = getattr(self, key)
            if not sub_value:
                setattr(self, key, super_value)
            else:
                # 字段结构进行拼接
                if isinstance(super_value, dict):
                    super_value = dict(super_value)
                    super_value.update(sub_value)
                    setattr(self, key, super_value)

                # 列表结构进行拼接
                if isinstance(super_value, list):
                    super_value = list(super_value)
                    super_value.extend(sub_value)
                    super_value = list(set(super_value))

                    setattr(self, key, super_value)

@dataclass
class EnvModel(BaseObject):
    _default = None

    def __init__(self, **kwargs):
        # 将数字类型转为字符串
        for k, v in kwargs.items():
            for clz in (int, float):
                if isinstance(v, clz):
                    v = str(v)
            kwargs[k] = v
        #  for k, v in dict(os.environ).items():
            #  setattr(self, k, v)
        super().__init__(**kwargs)

    @classmethod
    def default(cls):
        """使用默认环境变量

        """
        _env = {}
        if not cls._default:
            cls._default = cls(**_env)
        return cls._default

class FieldModel(PydanticModel):
    """配置字段
    在配置中，单一值无法描述字段时，可以使用该模型
    """

    doc: str = Field("", title="字段描述", alias="_doc")
    value: Any = Field("", title="值", alias="_value")
    data_type: Any = Field(None, title="类型", alias="_data_type")

    @validator('data_type')
    def format_data_type(cls, data_type):

        type_map = {
            "int": int,
            "str": str,
            "string": str,
            "float": float,
            "list": list,
            "dict": dict,
        }
        data_type = type_map.get(data_type, data_type)

        return data_type

    @root_validator
    def _root_validator(cls, values: dict):
        value = values.get('value')
        data_type = values.get('data_type')

        # 如果 data_type 为空，依照 value 来矫正
        if not data_type:
            data_type = type(value)
            values['data_type'] = data_type

        # format value
        values['value'] = cls._format_value(data_type, value)

        return values

    def __setattr__(self, key, value):
        if key == 'value':
            value = self._format_value(self.data_type, value)

        super().__setattr__(key, value)

    @staticmethod
    def _format_value(data_type, value):
        if not isinstance(value, data_type):
            if data_type in (int, str):
                value = data_type(value)
            if data_type in (list, dict):
                try:
                    value = json.loads(value)
                except:
                    raise ValueError(f"{value} is not {data_type}")
        return value

@dataclass
class AutoFieldModel():
    """该模型自动使用 FieldModel"""
    __annotations__ = {}

    def __init__(self, **kwargs):
        """对数据进行前置过滤"""
        for k, v in kwargs.items():
            setattr(self, k, v)

    def __setattr__(self, key, value):
        if key not in self.__annotations__:
            self.__annotations__[key] = FieldModel

        if not hasattr(self, key) or isinstance(value, FieldModel):
            value = self._format_value(value)
            super().__setattr__(key, value)
        else:
            print(key)
            origin_field = getattr(self, key)
            origin_field.value = value


    def dict(self):
        """重载方法
        value 值只保留 _value 部分
        """
        data = {}
        for key in self.__annotations__.keys():
            if not hasattr(self, key):
                continue
            data[key] = getattr(self, key).value
        return data

    def _format_value(self, v):
        if isinstance(v, dict):
            v = FieldModel(**v)
        if not isinstance(v, FieldModel):
            v = FieldModel(_value = v)
        return v


class RequestModel(PydanticModel, BaseModel):
    """请求配置模型"""

    name: str = Field(..., title="请求名称")
    title: str = Field('', title="请求标题")
    protocol: str = Field(ProtocolEnum.HTTP.value, title="请求协议")
    method: str = Field(MethodEnum.GET.value, title="请求方式")
    path: str = Field('', title="路径")
    url_prefix: str = Field('', title="地址前缀")
    url: str = Field('', title="地址")
    domain: str = Field('', title="域名")
    cookie_domains: List[str] = Field([], title="获取 cookie 的域名列表")
    cookies: Dict[str, Any] = Field({}, title="cookies 参数")
    headers: Dict[str, Any] = Field({}, title="headers 参数")
    params: AutoFieldModel = Field(AutoFieldModel(), title="地址参数")
    json_data: AutoFieldModel = Field(AutoFieldModel(), title="json 参数", alias="json")
    #  data: str = Field(None, title="域名")

    class Config:
        arbitrary_types_allowed = True

    @validator('protocol')
    def check_protocol(cls, v: str):
        ProtocolEnum.validate(v)
        return v

    @validator('method')
    def check_method(cls, v: str):
        v = v.upper()
        MethodEnum.validate(v)
        return v

    def add_params(self, **kwargs):
        """添加 params 参数"""
        for k, v in kwargs.items():
            setattr(self.params, k, v)

    def add_json(self, **kwargs):
        """添加 params 参数"""
        for k, v in kwargs.items():
            setattr(self.json_data, k, v)

    def dict(self):
        data = super().dict()
        data['params'] = self.params.dict()
        data.pop('json_data')
        data['json'] = self.json_data.dict()

        return data


class ModuleModel(PydanticModel, BaseModel):
    #  __req__: dict = {}

    name: str = Field(..., title="模块名称")
    protocol: str = Field(ProtocolEnum.HTTP.value, title="请求协议")
    domain: str = Field('', title="域名")
    url_prefix: str = Field('', title="地址前缀")
    cookie_domains: List[str] = Field([], title="获取 cookie 的域名列表")
    cookies: Dict[str, Any] = Field({}, title="cookies 参数")
    headers: Dict[str, Any] = Field({}, title="headers 参数")
    requests: List[Union[RequestModel, dict]] = Field([], title="请求列表")
    #  include = datatype.Str()

    class Config:
        arbitrary_types_allowed = True

    class Meta:
        request_map: Dict[str, Dict[str, RequestModel]
                ] = defaultdict(RequestModel)

    @validator('protocol')
    def check_protocol(cls, v: str):
        ProtocolEnum.validate(v)
        return v

    @validator('requests')
    def format_requests(cls, requests: list) -> List[RequestModel]:
        new_items = []
        for item in requests:
            if isinstance(item, dict):
                new_items.append(RequestModel(**item))
            if isinstance(item, RequestModel):
                new_items.append(item)
        #  cls.Meta.request_map[] = {o.name: o for o in new_items}
        return new_items

    @root_validator
    def _root_validator(cls, values):
        # 初始化 request_map
        name = values.get("name")
        cls.Meta.request_map[name] = {
            o.name: o for o in values.get("requests", [])}

        return values

    #  def __init__(self, *args, **kwargs):
        #  super().__init__(*args, **kwargs)
        #  self.__req__ = {o.name: o for o in self.requests}

    def get_request(self, name):
        """获取 request 模型"""
        #  req = self.__req__.get(name)
        #  print(self.Meta.request_map)
        req = self.Meta.request_map[self.name].get(name)
        if req:
            # 继承 module 的字段
            inherit_keys = ('protocol', 'domain', 'cookies', 'headers',
                'cookie_domains', 'url_prefix')
            req.inherit(self, inherit_keys)

            # 拼装 url
            if not req.url:
                req.url = f'{req.protocol}://{req.domain}' \
                    f'{self.url_prefix}{req.path}'
        return req


@dataclass(config = PydanticConfig)
class ConfigModel:
    """客户端全局配置模型"""
    __all__ = ['api_history_dir','server_port', 'server_host']
    modules: List[Union[ModuleModel, dict]] = Field([], title="模块列表")
    env: EnvModel = Field(EnvModel(), title="环境变量")
    cookies: Dict[str, Any] = Field({}, title="cookies 参数")
    headers: Dict[str, Any] = Field({}, title="headers 参数")
    cookie_domains: List[str] = Field([], title="获取 cookie 域名列表")
    modules_include: List[str] = Field([], title="模块加载列表")
    function_modules: List[str] = Field([], title="工具模块加载列表")
    server_port: str = Field(Constants.SERVER_PORT, title="服务默认端口")
    server_host: str = Field(Constants.SERVER_HOST, title="服务默认地址")
    api_history_dir: str = Field(Constants.API_HISTORY_DIR,
        title="请求历史记录存在目录")

    #  class Config:
        #  arbitrary_types_allowed = True

    class Meta:
        module_map: Dict[str, ModuleModel] = {}

    #  @validator('env')
    #  def format_env(cls, env: Union[EnvModel, dict]) -> EnvModel:
        #  print('=' * 100)
        #  print(env)
        #  if isinstance(env, dict):
            #  print(env)
            #  return EnvModel(**env)
        #  return env

    @validator('modules')
    def format_modules(cls, modules: List[Union[ModuleModel, dict]]
            ) -> List[ModuleModel]:
        new_modules = []
        for module in modules:
            if isinstance(module, dict):
                new_modules.append(ModuleModel(**module))
            if isinstance(module, ModuleModel):
                new_modules.append(module)
        cls.Meta.module_map = {o.name: o for o in new_modules}
        return new_modules

    @validator('function_modules')
    def format_function_modules(cls, modules: list) -> list:
        default_env = EnvModel.default()
        for i, _module in enumerate(modules):
            modules[i] = ConfigValue(_module
                ).set_env(**default_env.to_dict()).format()

        return modules

    def add_module(self, module: Union[ModuleModel, dict]):
        if isinstance(module, dict):
            module = ModuleModel(**module)
        self.modules.append(module)
        self.Meta.module_map[module.name] = module

    def get_module(self, name):
        """获取模块"""
        #  if not self.__is_format__:
            #  raise Exception('ConfigModel must format first')
        module = self.Meta.module_map.get(name)
        if module:
            inherit_keys = ['headers', 'cookies', 'cookie_domains']
            module.inherit(self, inherit_keys)

        return module

    @classmethod
    def _iter_path(cls, filepath):
        """迭代地址"""
        if os.path.isfile(filepath):
            yield filepath
        if os.path.isdir(filepath):
            names = os.listdir(filepath)
            for _dir, _, names in os.walk(filepath):
                for name in names:
                    yield os.path.join(_dir, name)
        yield None

    def iter_module_path(self):
        """迭代模块地址列表"""
        for module_path in self.modules_include or []:
            for path in self._iter_path(module_path):
                if not path:
                    continue
                # 过滤掉不是 yml 结尾的配置文件
                if not path.endswith('.yml'):
                    continue
                yield path




#!/usr/bin/env python
# -*- coding:utf-8 -*-
# Author: wxnacy@gmail.com
"""
run 命令的参数解析
"""
import os
import shutil

from csarg import CommandArgumentParserFactory
from csarg import Action

from wush.common.constants import Constants
from .command import CmdArgumentParser

@CommandArgumentParserFactory.register()
class ConfigArgumentParser(CmdArgumentParser):
    cmd = 'config'

    @classmethod
    def default(cls):
        """
        初始化一个实例
        """
        item = cls()
        item.add_argument('cmd')
        item.add_argument('--init', action=Action.STORE_TRUE.value,
            help="初始化配置文件")
        return item

    def run(self, text):
        args = self.parse_args(text)

        # 初始化配置文件
        if args.init:
            config_path = Constants.CONFIG_PATH
            if os.path.exists(config_path):
                print("配置文件已存在")
                return

            # 判断配置目录是否存在
            if not os.path.exists(Constants.CONFIG_DIR):
                os.mkdir(Constants.CONFIG_DIR)

            shutil.copyfile(Constants.get_sys_config_path(), config_path)
            print('生成配置文件:', config_path)



#!/usr/bin/env python
# -*- coding:utf-8 -*-
# Author: wxnacy(wxnacy@gmail.com)
"""
配置模块
"""
import os
import yaml
import copy
from wpy.functools import clock

from wush.common.config_value import ConfigValue
from wush.common.constants import Constants
from wush.common.loggers import get_logger
from wush.common.utils import load_module
from wush.config.models import ConfigModel
from wush.config.models import EnvModel
from wush.config.function import load_function, Function

__all__ = ['load_config']

logger = get_logger()

class Config(object):
    logger = get_logger('Config')

    _config = None
    _function: Function = None
    module_name = None
    space_name = None
    #  builtin_env = EnvModel()

    @classmethod
    def read_yml(cls, filepath):
        """读取 yml 文件"""
        with open(filepath, 'r') as f:
            return yaml.safe_load(f)

    @classmethod
    def load(cls, conf_file):
        """通过文件加载"""
        # 去读 yml 文件
        data = cls.read_yml(conf_file) or {}

        # 格式化 modules_include 地址
        config_dir = os.path.dirname(conf_file)
        full_modules_include = []
        for path in data.get("modules_include", []):
            path = os.path.join(config_dir, path)
            full_modules_include.append(path)
        data['modules_include'] = full_modules_include

        ins = cls.loads(data)
        # 获取配置目录
        #  ins._config_dir = os.path.dirname(conf_file)
        return ins

    @classmethod
    def loads(cls, data):
        """通过 dict 数据加载"""
        ins = cls()
        #  print(data)
        ins._config = ConfigModel(**data)

        # 对模块地址进行解析
        if not ins._config.modules:
            ins._config.modules = []
        for module_path in ins._config.iter_module_path():
            module = cls.read_yml(module_path)
            ins._config.add_module(module)

        # 加载插件模块
        for module_name in ins._config.function_modules:
            load_module(module_name)

        # 加载方法
        ins._function = load_function()

        for key in ConfigModel.__all__:
            setattr(ins, key, getattr(ins._config, key))

        return ins

    def add_env(self, key, value):
        """添加环境变量"""
        setattr(self._config.env, key, value)

    @property
    def env(self):
        return self._config.env.to_dict()

    @property
    def function(self) -> Function:
        return self._function

    def get_modules(self):
        """获取模块列表"""
        return self._config.modules

    def get_requests(self, module_name):
        """获取请求列表"""
        return self._config.get_module(module_name).requests

    @clock(fmt = Constants.CLOCK_FMT, logger_func = logger.info)
    def get_request(self, module_name, request_name, set_env=True, environs=None):
        """获取请求模型
        :param bool set_env: 是否设置环境变量
        :param dict environs: 编辑变量
        """
        req = self._config.get_module(module_name).get_request(request_name)
        req = copy.deepcopy(req)
        env = self._config.env.to_dict()
        if isinstance(environs, dict):
            env.update(environs)
        self.logger.info('config env {}'.format(env))

        for key, value in env.items():
            # 优先使用环境变量数据
            env_value = os.getenv(key)
            if env_value:
                env[key] = env_value
                continue
            if not value:
                continue
            os.environ[key] = value

        # 将请求模型做环境变量格式化处理并返回
        if set_env:
            req = ConfigValue(req).set_env(**env).set_functions(
                **self.function.get_functions()).format()
        return req


def _get_config_path(config_path=None):
    """
    获取配置地址
    """
    if config_path:
        Config.logger.info('配置文件: 用户指定 {}'.format(config_path))
        return config_path
    if os.path.exists(Constants.CONFIG_PATH):
        Config.logger.info('配置文件: 默认 {}'.format(Constants.CONFIG_PATH))
        return Constants.CONFIG_PATH

    config_path = Constants.get_sys_config_path()
    Config.logger.info('配置文件: 系统 {}'.format(config_path))
    return config_path

_config = None
_config_dict = {}

@clock(fmt = Constants.CLOCK_FMT, logger_func = logger.info)
def load_config(config_path = None):
    """加载配置
    :param str config_path: 配置文件路径

    1、优先使用用户指定配置文件
    2、否则使用默认配置文件路径
    3、否则使用初始化的配置

    """
    global _config
    if not _config:
        config_path = _get_config_path(config_path)
        _config = Config.load(config_path)

    return _config

#  if __name__ == "__main__":
    #  config = load_config()



#!/usr/bin/env python
# -*- coding:utf-8 -*-
# Author: wxnacy@gmail.com
"""

"""
#  import abc
from datetime import datetime
import pygments

#  from pygments.token import Token
from pygments.lexers.python import PythonLexer
from prompt_toolkit.formatted_text import PygmentsTokens
from prompt_toolkit import print_formatted_text

from csarg import CommandArgumentParser

from wpy.functools import clock
from wush.common.constants import Constants
from wush.common.loggers import create_logger
from wush.common.run_mode import RUN_MODE
from wush.config import load_config

class CmdArgumentParser(CommandArgumentParser):
    logger = create_logger('CmdArgumentParser')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


    def get_completions_after_argument(self, word_for_completion):
        """
        获取补全的单词列表
        :param word_for_completion: 补全需要的单词
        """
        return []

    def get_completions_after_cmd(self, argument, words=None):
        """获取补全使用的单词列表"""
        res = []
        if words and isinstance(words, list):
            res.extend(words)
        args = self.get_arguments()
        for arg in args:
            if arg.is_cmd:
                continue
            # 已经赋值的不需要展示
            # 列表除外
            if arg.is_set and not arg.is_list:
                continue
            res.append(dict(text = '--' + arg.name, display_meta = arg.help))
        return res

    def parse_args(self, text):
        """解析参数"""
        args = super().parse_args(text)

        # 输出日志
        arg_dict = {}
        arg_list = self.get_arguments()
        for arg in arg_list:
            #  log_text = f'{self.cmd} argument {arg.name}' \
                #  f' {getattr(args, arg.name.replace("-", "_"))}'
            arg_dict[arg.name] = getattr(args, arg.name.replace('-', '_'))
            #  self.logger.info(log_text)
        self.logger.info(f'{self.cmd} arguments {arg_dict}')
        return args

    def _print(self, text):
        tokens = list(pygments.lex(text, lexer=PythonLexer()))
        print_formatted_text(PygmentsTokens(tokens), end='')

    @clock(fmt = Constants.CLOCK_FMT, logger_func = logger.info)
    def run(self, args):
        begin_text = "cmd begin "
        self.logger.info(f"{begin_text:=<50}")
        self.config = load_config()

        # 添加历史记录
        if RUN_MODE.is_command:
            self.add_history(args)

        func_name = f'run_{RUN_MODE.mode}'
        func = None
        if hasattr(self, func_name):
            func = getattr(self, func_name)
        else:
            # 如果没有提供模式专用函数，则使用默认 run 方法
            if not hasattr(self, 'run'):
                raise RuntimeError('找不到参数运行方法')
            func = getattr(self, 'run')

        func(args)
        end_text = "cmd end "
        self.logger.info(f"{end_text:=<50}")

    @classmethod
    def add_history(cls, cmd):
        if isinstance(cmd, list):
            cmd = ' '.join(cmd)

        cmd.strip()

        if cmd.startswith('history'):
            return

        with open(Constants.HISTORY_PATH, 'r') as f:
            lines = f.readlines()

        historys = []
        for line in lines:
            if not line.startswith('+'):
                continue
            historys.append(line.strip('\n')[1:])

        if historys and historys[-1] == cmd:
            return

        lines = []
        lines.append(f"\n# {str(datetime.now())}")
        lines.append(f"\n+{cmd}\n")
        with open(Constants.HISTORY_PATH, 'a') as f:
            f.write(''.join(lines))




#!/usr/bin/env python
# -*- coding:utf-8 -*-
# Author: wxnacy@gmail.com
"""
run 命令的参数解析
"""

from csarg import Action
from csarg import CommandArgumentParserFactory
from inspect import getargspec

from wush.common.utils import list_key_val_to_dict
from wush.common.loggers import create_logger
from wush.config.function import load_function
from .command import CmdArgumentParser


@CommandArgumentParserFactory.register()
class FuncArgumentParser(CmdArgumentParser):
    cmd = 'func'
    logger = create_logger('FuncArgumentParser')

    @classmethod
    def default(cls):
        """
        初始化一个实例
        """
        item = cls()
        item.add_argument('cmd')
        item.add_argument('--doc', help='展示方法的文档信息')
        item.add_argument('--exec', help='执行方法')
        item.add_argument('--params', action = Action.APPEND.value,
            help='请求参数')
        return item

    def run_command(self, text):
        self.run_shell(text)

    def run_shell(self, text):
        functions = self.config.function.get_functions()
        args = self.parse_args(text)
        # 打印方法的文档
        if args.doc:
            for k, v in functions.items():
                if args.doc == k:
                    self._print(getattr(v, '__doc__'))
            return

        if args.exec:
            func_name = args.exec
            func = functions.get(func_name)
            params = list_key_val_to_dict(args.params or [])
            res = func(**params)
            print(res)
            return

        # 方法名的最大长度
        max_func_length = max([len(o) for o in functions.keys()])

        output = {
            'headers': [
                { "display": "方法名", 'width':max_func_length},
                { "display": "描述" },
                { "display": "模块" },
                ],
            'items': []
            }
        for k, v in functions.items():
            output['items'].append((
                k, self._get_short_doc(v), getattr(v, '__module__')))
        self.config.function.print_table(output)

    def _get_short_doc(self, func):
        """获取短文档"""
        doc = func.__doc__ or ''
        doc = doc.split('\n')[0]
        return doc[:20]

    def get_completions_after_argument(self, word_for_completion):
        """
        获取补全的单词列表
        :param word_for_completion: 补全需要的单词
        """
        words = []
        if not self.argument:
            return words
        arg = self.argument
        if word_for_completion == '--doc':
            # 自动补全返回方法名列表
            functions = load_function().get_functions()
            words = []
            for name, func in functions.items():
                words.append(dict( text = name,
                    display_meta=self._get_short_doc(func)))
            return words
        elif word_for_completion == '--exec':
            # 执行方法的自动补全
            functions = load_function().get_functions()
            words = []
            for name, func in functions.items():
                argspec = getargspec(func)
                arg_text = ', '.join(argspec.args)
                if argspec.varargs:
                    arg_text += f', *{argspec.varargs}'
                if argspec.keywords:
                    arg_text += f', **{argspec.keywords}'
                text = f'{name}({arg_text})'
                words.append(dict( text = name,
                    display = text,
                    display_meta=self._get_short_doc(func)))
            return words

        elif word_for_completion == '--params':
            # 参数
            functions = load_function().get_functions()
            params = {}
            for name, func in functions.items():
                if name != arg.exec:
                    continue
                argspec = getargspec(func)
                for arg_name in argspec.args:
                    params[arg_name] = ''
            words = self._dict_to_completions(params)
            return words

        return super().get_completions_after_argument(word_for_completion)

    def _dict_to_completions(self, data):
        """字典转为自动补全信息"""
        words = []
        for k, v in data.items():
            words.append(dict(
                text = '{}='.format(k),
                display = '{}={}'.format(k, v),
                display_meta=''))
        return words



#!/usr/bin/env python
# -*- coding:utf-8 -*-
# Author: wxnacy(wxnacy@gmail.com)
"""

"""
#  import json

from wpy.base import BaseEnum

class ModeEnum(BaseEnum):
    COMMAND = 'command'
    SHELL = 'shell'

class RunMode:
    #  name = 'run_mode'
    _mode = None

    def __init__(self, mode = None):
        self._mode = mode

    @property
    def mode(self):
        """模式"""
        return self._mode

    @property
    def is_command(self):
        """是否为命令模式"""
        return self._mode == ModeEnum.COMMAND.value

    @property
    def is_shell(self):
        """是否为 shell 模式"""
        return self._mode == ModeEnum.SHELL.value

    def set_command(self):
        self._mode = ModeEnum.COMMAND.value

    def set_shell(self):
        self._mode = ModeEnum.SHELL.value

    #  def dict(self):
        #  return { "mode": self.mode }

    #  def json(self):
        #  return json.dumps(self.dict())


RUN_MODE = RunMode()



#!/usr/bin/env python
# -*- coding:utf-8 -*-
# Author: wxnacy@gmail.com
"""

"""

import json
import os
import re
from functools import singledispatch

from wpy.path import read_dict
from wush.common.loggers import create_logger
from wush.model import Model

__all__ = ['environ_keys', 'ConfigValue']

#  _REG_ENV = r'(\${.*?})'
#  _REG_ENV = r'\{(.+?)\}'
_REG_ENV2 = r'\${(.+?)\}'

@singledispatch
def environ_keys(obj):
    """获取环境变量 keys"""
    if not hasattr(obj, '__annotations__'):
        return
    res = set()
    for key in obj.__annotations__.keys():
        env_keys = environ_keys(getattr(obj, key))
        for _key in env_keys:
            res.add(_key)

    return res

@environ_keys.register(str)
def _(text):
    return set(re.findall(_REG_ENV2, text))

@environ_keys.register(dict)
def _(data):
    res = set()
    for key, value in data.items():
        for name in environ_keys(key):
            if name not in res:
                res.add(name)
        for name in environ_keys(value):
            if name not in res:
                res.add(name)
    return res

@environ_keys.register(list)
def _(data):
    res = set()
    for line in data:
        for name in environ_keys(line):
            if name not in res:
                res.add(name)
    return res

@environ_keys.register(Model)
def _(model):
    res = set()
    for value in model.dict().values():
        for name in environ_keys(value):
            if name not in res:
                res.add(name)
    return res

class ConfigValue():
    logger = create_logger('ConfigValue')

    def __init__(self, value):
        self.value = value

        #  self.REG_ENV = r'(\${.*?})'
        #  self.REG_ENV2 = r'({.*?})'
        self.REG_PARSE = r'(^\b(json|yml|xml)\b@.*?)'
        self.env = dict(os.environ)
        self.functions = {}

    def set_env(self, **data):
        self.env.update(data)
        return self

    def set_functions(self, **data):
        self.functions.update(data)
        return self

    def format(self):
        '''输出格式化信息'''
        return self._format(self.value)

    def _format(self, value):
        if not value:
            # 为空，返回
            return value

        #  格式化响应格式数据
        for t in (str, list, dict):
            if isinstance(value, t):
                func_name = '_format_' + t.__name__
                return getattr(self, func_name)(value)

        # 格式化模型
        if isinstance(value, Model):
            for k, v in value.to_dict().items():
                v = self._format(v)
                setattr(value, k, v)
            return value

        #  if isinstance(value, BaseObject):
            #  for k, v in value.to_dict().items():
                #  v = self._format(v)
                #  setattr(value, k, v)

        self._format_object(value)
        return value

    def _format_object(self, obj):
        if not hasattr(obj, '__annotations__'):
            return
        for key in obj.__annotations__.keys():
            if not hasattr(obj, key):
                continue
            val = ConfigValue(getattr(obj, key)).set_env(**self.env
                ).set_functions(**self.functions).format()
            setattr(obj, key, val)

    def _format_dict(self, value):
        """格式化字典"""
        for v_k, v_v in value.items():
            value[v_k] = ConfigValue(v_v).set_env(**self.env
                    ).set_functions(**self.functions).format()
        return value

    def _format_list(self, lines):
        """格式化数组"""
        for i in range(len(lines)):
            lines[i] = ConfigValue(lines[i]).set_env(**self.env
                    ).set_functions(**self.functions).format()
        return lines

    def _format_str(self, text):
        '''格式化字符串'''
        if self._has_environs(text):
            text = self._format_environ(text)
        if text.startswith('@'):
            text = text[1:]
            #  self.logger.debug(text)
            if text.startswith('{') and text.endswith('}'):
                return json.loads(text)
            if text.startswith('[') and text.endswith(']'):
                return json.loads(text)

            if not self._is_file(text):
                raise ValueError(u'{} 文件路径不存在'.format(text))

            return self._format_file_content(text)

        parse_type = self._get_parse_type(text)
        if parse_type:
            text = text[len(parse_type) + 1:]
            if self._is_file(text):
                if parse_type == 'json':
                    return read_dict(text)
            else:
                if parse_type == 'json':
                    return json.loads(text)

        return text

    def _get_parse_type(self, text):
        """获取解析类型"""
        lines = re.findall(self.REG_PARSE, text)
        if not lines:
            return None
        return lines[0][-1]

    def _format_file_content(self, filepath):
        '''格式化文件内容'''
        if filepath.endswith('.json') or filepath.endswith('.yml'):
            return read_dict(filepath)

        with open(filepath, 'r') as f:
            lines = f.readlines()
            if not lines:
                return ''
            if lines[0].startswith('{') and lines[-1].endswith('}'):
                return read_dict(filepath)
            return ''.join(lines)

    def _is_file(self, filepath):
        '''是否为文件'''
        return os.path.exists(filepath)

    def _has_environs(self, text):
        '''是否有环境变量'''
        if not isinstance(text, str):
            return False
        self.environ_names = environ_keys(text)
        return True if self.environ_names else False

    def _replace(self, match):
        """正则替换方法"""
        groups = match.groups()
        if not groups:
            return

        k = groups[0]
        if '(' not  in k and ')' not in k:
            env_value = self.env.get(k, "")
            if isinstance(env_value, dict) or isinstance(env_value, list):
                return json.dumps(env_value)
            return str(env_value)

        k = k.strip(')').strip(' ')
        func_name, args_str = k.split('(')
        func = self.functions.get(func_name)
        #  self.logger.info(f'func_name {func_name} {func}')

        if not func:
            return
        if not args_str:
            args_str = '()'
        else:
            args_str = '({},)'.format(args_str)
        args = eval(args_str)
        repl = func(*args)
        repl = str(repl)
        return repl

    def _format_environ(self, text):
        '''格式化环境变量'''
        if not self.environ_names:
            return text

        text = re.sub(_REG_ENV2, self._replace, text)
        #  text = re.sub(_REG_ENV, self._replace, text)
        return text

if __name__ == "__main__":
    text_value = {"name": "wxnacy"}
    text = 'json@${text_value}'
    data = ConfigValue(text).set_env(text_value = text_value).format()
    assert data == {"name": "wxnacy"}



#!/usr/bin/env python
# -*- coding:utf-8 -*-
# Author: wxnacy(wxnacy@gmail.com)
"""
配置可用的方法模块
"""

import traceback
import os
import hashlib
import json
from typing import (
    Dict, Callable
)

from rich.console import Console
from rich.table import Table
from wpy.base import BaseFactory
from wpy import randoms

from wush.common import utils
from wush.common.loggers import get_logger
from wush.web.cookie import Cookie

logger = get_logger('function')

__all__ = ['FunctionFactory']

class FunctionFactory(BaseFactory):

    @classmethod
    def get_super_factory(cls):
        factory = cls.get_factory()
        _super_factory = {}
        for key, value in factory.items():
            module = value.__module__
            if module.startswith('wush'):
                _super_factory[key] = value
        return _super_factory

@FunctionFactory.register()
def random_int(length: int, min_int: int=None, max_int: int=None):
    """随机 int 值"""
    return randoms.random_int(length, min_int, max_int)

@FunctionFactory.register()
def random_str(length: int, source=None):
    """随机 int 值"""
    return randoms.random_str(length, source)

@FunctionFactory.register()
def md5(text):
    """计算 md5 摘要"""
    return hashlib.md5(text.encode('utf-8')).hexdigest()

@FunctionFactory.register()
def get_completion_words(word_for_completion):
    """获取补全的单词列表"""
    return []

@FunctionFactory.register()
def handler_response(response):
    #  print('Response:')
    # 判断是否为 html 结果
    try:
        data = response.json()
        data = json.dumps(data, indent=4, ensure_ascii=False)
    except:
        logger.error(traceback.format_exc())
        data = response.content
    print(data)
    url = response.url
    if response.is_html and not response.request_builder.argument.no_browser:
        print('Html page see in browser')
        utils.open_url(url)

@FunctionFactory.register()
def print_table(config):
    """打印表格
    :param dict config: 表格数据
        结构如下：
        {
            "headers": [
                { "display": "列名", 'width': '列宽度，非必传' }
            ],
            "items": [
                ('列数据, 长度需要和 headers 长度保持一致')
            ]
        }
    """
    console = Console()
    headers = config.get("headers") or []
    table = Table(show_header=True, show_lines=True,
        header_style="bold magenta")
    for header in headers:
        name = header.pop("display", '未命名')
        table.add_column(name, **header)
    items = config.get("items", [])
    for item in items:
        table.add_row(*item)
    console.print(table)
    after_table = config.get("after_table")
    if after_table:
        console.print(after_table)

@FunctionFactory.register()
def get_browser_cookies(domains, cookie_name=None):
    """
    获取浏览器的 cookies
    """
    res = Cookie.get_browser_cookie(*domains) or {}
    if cookie_name:
        return res.get(cookie_name)
    return res

class Function(object):
    get_completion_words = None
    random_int = None
    random_str = None
    handler_response = None
    test = None

    _functions: Dict[str, Callable] = {}

    def __init__(self, functions):
        self._functions = functions
        for name, func in functions.items():
            if isinstance(func, str):
                raise Exception('func {} can not be str'.format(name))
            setattr(self, name, func)

    def add_function(self, func: Callable):
        """添加方法"""
        setattr(self, func.__name__, func)
        self._functions[func.__name__] = func

    def get_functions(self) -> Dict[str, Callable]:
        """获取方法字段"""
        return self._functions

_super_function = None

def load_super_function():
    """加载 super function"""
    global _super_function
    if not _super_function:
        _super_function = Function(FunctionFactory.get_super_factory())
    return _super_function


_function = None
def load_function():
    """加载全部 function"""
    global _function
    if not _function:
        _function = Function(FunctionFactory.get_factory())
    return _function

if __name__ == "__main__":
    pass



#!/usr/bin/env python
# -*- coding:utf-8 -*-
# Author: wxnacy(wxnacy@gmail.com)
"""

"""
import json

from wush.common.utils import run_shell
from wush.common.loggers import get_logger
from wush.config import load_config
from wush.web.request import RequestClient
from wush.web.request import RequestBuilder

logger = get_logger('wush.run')

__all__ = ['run']

def run(module_name, request_name, **kwargs):
    params = kwargs.get("params", {})
    env = kwargs.get("env", {})
    json_data = kwargs.get("json", {})
    config_path = kwargs.get("config")
    with_browser_cookie = kwargs.get("with_browser_cookie", False)

    config = load_config(config_path)
    request_model = config.get_request(module_name, request_name,
        environs = env)
    request_model.add_params(**params)
    request_model.add_json(**json_data)

    builder = RequestBuilder.loads_request_model(request_model,
        with_browser_cookie)

    req_client = RequestClient(builder)
    return req_client.request()


def run_in_shell(module_name, request_name, **kwargs):
    params = kwargs.get("params", {})
    env = kwargs.get("env", {})
    json_data = kwargs.get("json", {})
    config = kwargs.get("config")
    cmd = f'wush run --module {module_name} --name {request_name}'
    if params:
        for key, value in params.items():
            cmd += f' --params {key}={value}'
    if env:
        for key, value in env.items():
            cmd += f' --env {key}={value}'
    if json_data:
        for key, value in json_data.items():
            cmd += f' --json {key}={value}'
    if config:
        cmd += f' --config {config}'
    logger.info(f'run cmd {cmd}')
    res = run_shell(cmd)
    lines = []
    for line in res:
        line = line.decode()
        for line in line.split('\n'):
            if line.startswith('('):
                continue
            line = line.replace('\n', '')
            lines.append(line)
    text = ''.join(lines)
    logger.info(f'run res {text}')

    return json.loads(text)


if __name__ == "__main__":
    res = run('wush', 'test_get', params={"age": 1})
    print(res.json())





#!/usr/bin/env python
# -*- coding:utf-8 -*-
# Author: wxnacy@gmail.com
"""

"""
import os
import sys

from flask import Flask
from flask import request

from wush.common.loggers import create_logger
from wush.web.utils import telnet
from wush.web.history import History
from wush.config import load_config

app = Flask(__name__)
client = None

logger = create_logger(__name__)

@app.route('/api/version/<string:version>')
def detail(version):
    res =  History().read(version)
    log_text = f'{request.path} {res}'
    logger.info(log_text)
    return res

@app.route('/api/version/<string:version>/<string:type>')
def detail_type(version, type):
    data = History().read(version)
    res = data.get(type, {})
    if res.get("is_json"):
        return res.get("json")
    else:
        return res.get("text")

@app.route('/api/callback')
def get_callback():
    config = load_config()
    params = dict(request.args)
    func_name = params.pop("func", None)
    if not func_name:
        return { "func": None }

    func = config.function.get_functions().get(func_name)
    if not func:
        return { "func": None }

    return func(**params)

@app.route('/api/test', methods=['post', 'get'])
def test():
    res = {
        "args": request.args,
        "json": request.json,
        "data": str(request.data),
        "headers": dict(request.headers),
        "cookies": dict(request.cookies)
    }
    return res

def clear_stdout():
    # 刷新缓冲区
    sys.stdout.flush()
    sys.stderr.flush()

    # 重定向标准输入、输出、错误的描述符
    # dup2函数原子化地关闭和复制文件描述符，重定向到/dev/nul，即丢弃所有输入输出
    with open('/dev/null') as read_null, \
            open('/dev/null', 'w') as write_null:
        os.dup2(read_null.fileno(), sys.stdin.fileno())
        os.dup2(write_null.fileno(), sys.stdout.fileno())
        os.dup2(write_null.fileno(), sys.stderr.fileno())

def run_server(clear_logger: bool = True, config_path: str = None):
    # 清空输出
    if clear_logger:
        clear_stdout()
    # 获取配置
    config = load_config(config_path)
    port = int(config.server_port)
    host = config.server_host
    # 如果端口还没有启动，则启动服务
    if telnet(host, port):
        print('服务已经启动')
        return
    app.run(host = host, port=port)

if __name__ == "__main__":
    run_server()



#!/usr/bin/env python
# -*- coding:utf-8 -*-
# Author: wxnacy@gmail.com
"""
run 命令的参数解析
"""
from csarg import CommandArgumentParserFactory

from wush.common.constants import Constants
from wush.common.loggers import create_logger
from .command import CmdArgumentParser


@CommandArgumentParserFactory.register()
class HistoryArgumentParser(CmdArgumentParser):
    cmd = 'history'
    logger = create_logger()

    @classmethod
    def default(cls):
        """
        初始化一个实例
        """
        item = cls()
        item.add_argument('cmd')
        return item

    def run_shell(self, text):
        items = self._prompt.history.get_strings()
        self._print_historys(items)

    def run_command(self, text):
        with open(Constants.HISTORY_PATH, 'r') as f:
            lines = f.readlines()

        historys = []
        for line in lines:
            if not line.startswith('+'):
                continue
            historys.append(line.strip('\n')[1:])

        self._print_historys(historys)

    def _print_historys(self, historys: list):
        history_max_num_len = len(str(len(historys)))
        for i, item in enumerate(historys):
            show_index = i + 1
            show_index_fmt = '{{:<{}d}}'.format(history_max_num_len)
            line = '{} {}'.format(show_index_fmt.format(show_index), item)
            self._print(line)




#!/usr/bin/env python
# -*- coding:utf-8 -*-
# Author: wxnacy@gmail.com
"""
命令补全
"""

from prompt_toolkit.completion import Completion

from .base import BaseCompleter

__all__ = ['WordCompleter']

class WordCompleter(BaseCompleter):

    def __init__(self, words):
        self.words = words

    def get_completions(self, document, complete_event):
        super().get_completions(document, complete_event)
        for compl in self.search(self.words):
            start_position = self.get_start_position()
            # 如果列表是补全类，添加位置直接返回
            if isinstance(compl, Completion):
                compl.start_position = start_position
            else:
                compl = Completion(compl, start_position=start_position)
            yield compl




#!/usr/bin/env python
# -*- coding:utf-8 -*-
# Author: wxnacy@gmail.com
"""
run 命令的参数解析
"""
from csarg import CommandArgumentParserFactory

from wush.common.loggers import create_logger
#  from wush.models import Version
from .command import CmdArgumentParser


@CommandArgumentParserFactory.register()
class ViewArgumentParser(CmdArgumentParser):
    cmd = 'view'
    logger = create_logger('ViewArgumentParser')
    _versions = []

    @classmethod
    def default(cls):
        """
        初始化一个实例
        """
        item = cls()
        item.add_argument('cmd')
        item.add_argument('--grep', help='过滤')
        item.add_argument('--open', help='打开')
        item.add_argument('--page', help='页数', datatype=int, default=1)
        item.add_argument('--size', help='每页条数', datatype=int, default=10)
        item.add_argument('--version', help='请求版本')
        return item

    def run(self, text):
        args = self.parse_args(text)
        if args.open:
            self._print('See in browser')
            open_version(args.open)
            return
        self._list(args)

    def _list(self, args):
        items = Version.find()
        items.sort(key = lambda x: x._id, reverse=True)
        data = {
            'headers': [
                { "display": "Version" },
                { "display": "Space" },
                { "display": "Module" },
                { "display": "Name" },
                ],
            'items': []
            }

        def _filter(x):
            text = args.grep
            if text in x.space_name or text in x.module_name or \
                    text in x.request_name:
                return True
            return False

        # 过滤列表
        if args.grep:
            items = list(filter(_filter, items))

        total_count = len(items)

        # 分页展示
        start = (args.page - 1) * args.size
        end = start + args.size
        items = items[start:end]
        self._versions = items
        for item in items:
            data['items'].append((item._id, item.space_name, item.module_name,
                item.request_name))
        if end < total_count:
            data['after_table'] = 'With `--page {}` to see more'.format(args.page + 1)
        self.config.function.print_table(data)

    def get_completions_after_argument(self, word_for_completion):
        """
        获取补全的单词列表
        :param word_for_completion: 补全需要的单词
        """
        words = []
        if not self.argument:
            return words
        arg = self.argument
        if word_for_completion == '--open':
            items = self._versions
            if not items:
                items = Version.find()
                items.sort(key = lambda x: x._id, reverse=True)
            for item in items:
                dm = '{}_{}_{}'.format(item.space_name, item.module_name,
                    item.request_name)
                words.append(dict( text = item._id, display_meta=dm))
            return words

        return super().get_completions_after_argument(word_for_completion)



#!/usr/bin/env python
# -*- coding:utf-8 -*-
"""

"""

import requests
import json
from datetime import datetime
from typing import (
    Dict, Any
)
from pydantic import (
   BaseModel, Field, validator
)

from wpy.functools import clock

from wush.common.constants import Constants
from wush.common.loggers import get_logger
from wush.common.run_mode import RunMode
from wush.web.curl_utils import cUrl
from wush.web.enums import MethodEnum
from wush.web.enums import RequestsParamsEnum
from wush.web.response import ResponseClient
from csarg.parser import Argument

__all__ = ['RequestBuilder']

def create_version() -> str:
    return datetime.now().strftime('%Y%m%d%H%M%S_%s')

class RequestBuilder(BaseModel):
    """请求构造器"""
    version: str = Field(title="版本号", default_factory = create_version)
    url: str = Field(..., title="地址")
    method: str = Field(MethodEnum.GET.value, title="请求方式")
    params: Dict[str, Any] = Field({}, title="地址参数")
    json_data: Dict[str, Any] = Field({}, title="json 参数", alias="json")
    body: str = Field(None, title="body 参数")
    headers: Dict[str, Any] = Field({}, title="headers 参数")
    cookies: Dict[str, Any] = Field({}, title="cookies 参数")
    run_mode: RunMode = Field(None, title="运行模式")
    argument: Argument = Field(None, title="参数解析器")

    class Config:
        arbitrary_types_allowed = True

    @validator('method')
    def check_and_format_method(cls, v: str) -> str:
        v = v.upper()
        MethodEnum.validate(v)
        return v

    @classmethod
    def load_curl(cls, curl_file):
        """加载 curl 的本文文件"""
        params = cUrl.dump(curl_file)
        ins = cls(**params)
        ins.format()
        return

    def add_headers(self, **kwargs):
        """添加 headers"""
        self.headers.update(kwargs)

    def add_cookies(self, **kwargs):
        """添加 cookies"""
        self.cookies.update(kwargs)

    def set_run_mode(self, mode):
        self.run_mode = RunMode(mode)

    def to_requests(self):
        """转换为 requests 参数"""
        data = {}
        for key in RequestsParamsEnum.values():
            try:
                data[key] = getattr(self, key)
            except:
                pass
        data['json'] = self.json_data
        return data


class RequestClient:
    """请求客户端"""
    logger = get_logger('RequestClient')

    def __init__(self, builder: RequestBuilder):
        self.builder = builder

    def request(self):
        """发送请求"""
        params = self.builder.to_requests()
        self.logger.info(f'request builder {json.dumps(params, indent=4)}')
        res = self._request(**params)
        res_client = ResponseClient(self.builder, res)
        if res_client.status_code in (301, 302):
            url = res_client.location
            res = self._request(url = url)
            res_client = ResponseClient(self.builder, res)

        return res_client

    @clock(fmt = Constants.CLOCK_FMT, logger_func = logger.info)
    def _request(self, **params):
        return requests.request(**params)




#!/usr/bin/env python
# -*- coding:utf-8 -*-
# Author: wxnacy(wxnacy@gmail.com)
# Description:

from functools import singledispatch
from functools import lru_cache

def filter(source, *args, **kwargs):
    source_include = args if args else kwargs.get('source_include')
    source_exclude = kwargs.get('source_exclude')
    return DictFilter(source).filter(source_include = source_include,
        source_exclude = source_exclude)

@singledispatch
def _filter(data, include=[], exclude=[]):
    print('_filter')

@_filter.register(dict)
def _filter_dict(data, include=[], exclude=[]):
    print('_filter_dict')

@_filter.register(list)
def _filter_list(data, **kw):
    for d in data:
        _filter_dict(data, **kw)

class DictFilter(dict):
    def filter(self, *args, **kwargs):
        """
        过滤dict
        :param args: 默认 source_include
        :param kwargs:
            source_include：想要留下的keys
                eq:[attr,attr1]
                子元素可以使用 ["obj.attr"] 和 ["obj[attr1,attr2]"] 两种方式
                速度上推荐使用 ["obj[attr1,attr2]"]
            source_exclude：想要去掉的keys
        :return:
        """
        source_include = args if args else kwargs.get('source_include')
        source_exclude = kwargs.get('source_exclude')

        def _filter(t, o, k):
            """
            过滤key
            :param t: 过滤结果
            :param o: 目标对象
            :param k: 需要过滤的key
            :return:
            """
            if k in o:
                t[k] = o[k]
            return t

        def _kv1(k):
            key = k.split('.', 1)[0]
            sub_key = k.split('.', 1)[1]
            return key, sub_key

        def _kv2(k):
            key = k.split('[')[0]
            sub_keys = k.split('[')[1].rstrip(']').split(',')
            return key, sub_keys

        def _include(t, o, k):
            """
            判断key需要何种过滤
            :param t: 过滤结果
            :param o: 目标对象
            :param k: 需要过滤的key
            :return:
            """
            if '.' in k:
                key, sub_key = _kv1(k)
                v = o.get(key)
                if o.get(key) and isinstance(o.get(key), dict):
                    if key not in t:
                        t[key] = {}
                    t[key] = _include(t[key], o[key], sub_key)
                elif v and isinstance(v, list):
                    if key not in t:
                        t[key] = []
                    for i in range(len(v)):
                        temp = DictFilter(v[i]).filter(sub_key)
                        if len(t[key]) == i:
                            t[key].append(temp)
                        else:
                            t[key][i].update(temp)

            elif '[' in k and ']' in k:
                key, sub_keys = _kv2(k)
                v = o.get(key)
                if v and isinstance(v, dict):
                    t[key] = DictFilter(v).filter(*sub_keys)
                elif v and isinstance(v, list):
                    t[key] = [DictFilter(sv).filter(*sub_keys) for sv in v]
            else:
                t = _filter(t, o, k)

            return t

        def _exclude(o, k):
            if '.' in k:
                key, sub_key = _kv1(k)
                v = o.get(key)
                if v and isinstance(v, dict):
                    if sub_key in o[key]:
                        o[key].pop(sub_key)
                elif v and isinstance(v, list):
                    o[key] = [DictFilter(o).filter(source_exclude=[sub_key])
                            for o in v]

            elif '[' in k and ']' in k:
                key, sub_keys = _kv2(k)
                v = o.get(key)
                if v and isinstance(v, dict):
                    o[key] = DictFilter(v).filter(source_exclude=sub_keys)
                elif v and isinstance(v, list):
                    o[key] = [DictFilter(o).filter(source_exclude=sub_keys)
                            for o in v]
            else:
                if k in o:
                    o.pop(k)

        temp = {}
        if source_include:
            for item in source_include:
                temp = _include(temp, self, item)
            return temp
        elif source_exclude:
            for item in source_exclude:
                _exclude(self, item)
            return self
        return self

    def __getattr__(self, item):
        return self[item]


if __name__ == "__main__":
    #  print(filter(dict(id = 1, name = 'wxnacy', age = 12), 'id', 'name', exclude=['name']))
    #  print(filter([dict(id = 1, name = 'wxnacy', age = 12)] * 3, 'id', 'name', exclude=['name']))
    #  print(filter(dict(book=dict(id = 1, name = 'wxnacy', age = 12)), 'book.id',
        #  'book.name', exclude=['name']))
    data = [{ "id": 1 }]
    _filter(data)
    data = { "id": 1 }
    _filter(data)



#!/usr/bin/env python
# -*- coding:utf-8 -*-
"""

"""

from routes import Mapper
from wpy.base import BaseObject

route_map = Mapper()
route_map.connect(None, "/api/pg/test?id={id}", controller="pangu", action='test')

#  res = route_map.match('/api/pg/test')
#  print(res)

class Route(BaseObject):

    def __init__(self, url, endpoint, methods=None):
        super().__init__(url = url, endpoint = endpoint, methods=methods)
        if not self.methods:
            self.methods = ['GET']

class RouteGroup(BaseObject):

    def __init__(self, name, url_prefix=None):
        super().__init__(name = name, url_prefix = url_prefix)
        self._routes = []

    def route(self,url, methods=None):
        def decorate(func):
            r = Route(url, func)
            self._routes.append(r)
            return func
        return decorate

    def get_routes(self):
        return self._routes

class RouteFactory(BaseObject):
    _routes_endpoints = {}
    _route_map = Mapper()

    def _add_route(cls, route):
        key = route.endpoint.__name__
        if key in cls._routes_endpoints:
            raise Exception('endpoint {} is exists'.format(key))
        cls._routes_endpoints[key] = route

    def register_route(cls, route):
        if isinstance(route, Route):
            cls._add_route(route)
        if isinstance(route, RouteGroup):
            for r in route.get_routes():
                cls._add_route(r)

        for k, route in cls._routes_endpoints.items():
            cls._route_map.connect(None, r.url, controller='', action=k)

    def match(url, method='GET'):
        """匹配"""


factory = {}

def register(url):
    def decorate(func):
        key = func.__name__
        print(func.__module__)
        print(func.__name__)
        print(func.__new__)
        print(dir(func))
        route_map.connect(None, url, controller="test", action=key)
        factory[key] = func
        return func
    return decorate


#  class Test():

rg = RouteGroup('test')

@rg.route('/test')
def test():
    print('test')

print(rg.get_routes())



#!/usr/bin/env python
# -*- coding:utf-8 -*-
# Author: wxnacy@gmail.com
"""
run 命令的参数解析
"""
import json
from csarg import Action
from csarg import CommandArgumentParserFactory

from wush.common.loggers import create_logger
from wush.web.cookie import Cookie

from .command import CmdArgumentParser


@CommandArgumentParserFactory.register()
class CookieArgumentParser(CmdArgumentParser):
    cmd = 'cookie'
    logger = create_logger('CookieArgumentParser')

    @classmethod
    def default(cls):
        """
        初始化一个实例
        """
        item = cls()
        item.add_argument('cmd')
        item.add_argument('--domain', action = Action.APPEND.value,
            help='域名')
        item.add_argument('--to-str', action = Action.STORE_TRUE.value,
            help='是否转为字符串')

        return item

    def get_completions_after_argument(self, word_for_completion):
        """
        获取补全的单词列表
        :param word_for_completion: 补全需要的单词
        """
        words = []
        if not self.argument:
            return words
        #  arg = self.argument
        #  if word_for_completion == '--name':
            #  # 针对 --name 参数的自动补全
            #  requests = self.config.get_requests(arg.module)
            #  words = []
            #  for req in requests:
                #  words.append(dict( text = req.name, display_meta=req.title))
            #  return words

        return super().get_completions_after_argument(word_for_completion)

    def run_command(self, text):
        """运行命令行模式"""
        args = self.parse_args(text)
        if args.domain:
            res = Cookie.get_browser_cookie(*args.domain)
            if args.to_str:
                lines = [f'{k}={v};' for k, v in res.items()]
                print(' '.join(lines))

            else:
                print(json.dumps(res, indent=4))
            return
        print('使用 --domain 指定域名')

    def run_shell(self, text):
        self.run_command(text)




#!/usr/bin/env python
# -*- coding:utf-8 -*-
# Author: wxnacy(wxnacy@gmail.com)
"""
cookie 相关模块
"""
import traceback

from wpy.functools import clock
from wush.common.loggers import get_logger
from wush.common.constants import Constants

class Cookie(object):

    logger = get_logger('Cookie')

    @classmethod
    @clock(fmt = Constants.CLOCK_FMT, logger_func = logger.info)
    def get_browser_cookie(cls, *domains, browser='chrome'):
        """获取留来的 cookie
        :param list domains: 域名列表
        """
        chrome = None
        try:
            import browsercookie
            chrome = getattr(browsercookie, browser)()
        except :
            print('can not import browsercookie')
            traceback.print_exc()
            traceback.print_stack()
        res = {}
        if not chrome:
            return res
        try:
            for k, v in chrome._cookies.items():
                if k not in domains:
                    continue
                cookies = v.get("/")
                for ck, cv in cookies.items():
                    res[ck] = cv.value
        except Exception as e:
            print(e)
        return res

if __name__ == "__main__":
    res = Cookie.get_browser_cookie('.python.org')
    print(res)




#!/usr/bin/env python
# -*- coding:utf-8 -*-
# Author: wxnacy@gmail.com
"""

"""
from datetime import datetime
from loguru import logger

LOG_FILE = '/tmp/wush.log'
LOG_FORMAT = '[<g>{time:HH:mm:ss.SSS}</g>][<y>{elapsed}</y>]' \
    '[<level>{level}</level>][<c>{name}:{function}:{line}</c>] ' \
    '<level>{message}</level>'

def format_logger(record):
    #  print(record)
    return record


def make_const_extra():
    extra = {}
    #  extra.setdefault('last_time', datetime.now())
    return extra

def patch_record(record):
    #  print(record)

    # 简化包路径名
    name_l = record['name'].split('.')
    if len(name_l) > 2:
        name_l[:-1] = [w[0] for w in name_l[:-1]]
    record.update(
        name='.'.join(name_l),
    )

    # 时间消耗
    record['elapsed'] = str(record['elapsed'])[2:-3]


logger.remove()
#  with logger.contextualize(start_time = datetime.now()):
logger.configure(extra = make_const_extra(), patcher = patch_record)
logger.add(
    LOG_FILE, format=LOG_FORMAT, colorize=True, rotation="50 MB", retention=5
)
#  logger.add( sys.stdout, format = LOG_FORMAT,)

#  handlers = [
        #  {"sink": str(file), "format": "FileSink: {message}"},
        #  {"sink": sys.stdout, "format": "StdoutSink: {message}"},
    #  ]
#  logger.configure(handlers=handlers)


def create_logger(name=None):
    return logger

def get_logger(name=None):
    return create_logger(name=None)


if __name__ == "__main__":
    import time
    logger.debug(f"debug {datetime.now()}")
    time.sleep(1)
    logger.info("info %s", datetime.now())



#!/usr/bin/env python
# -*- coding:utf-8 -*-
# Author: wxnacy(wxnacy@gmail.com)
"""

"""

from wpy.base import BaseEnum

class BasetypeEnum(BaseEnum):
    STR = str
    INT = int
    FLOAT = float
    LIST = list
    DICT = dict



#!/usr/bin/env python
# -*- coding:utf-8 -*-
# Author: wxnacy@gmail.com
"""
命令补全
"""

from prompt_toolkit.completion import Completer
from prompt_toolkit.completion import Completion

from wpy.lists import search

__all__ = [
    "BaseCompleter"
]

class BaseCompleter(Completer):
    char_before_cursor = ''
    current_line_before_cursor = ''
    first_word = ''

    def yield_completer(self, completer):
        items = completer.get_completions(self.document, self.complete_event)
        for item in items:
            yield item

    @property
    def first_word(self):
        """
        获取第一个单词
        """
        def _():
            current_line_before_cursor = self.current_line_before_cursor.strip()
            args = current_line_before_cursor.split(' ')
            args_len = len(args)
            return args[0] if args_len > 0 else ''
        res = _()
        return res

    @property
    def word_for_completion(self):
        """补全使用的单词"""
        def _():
            char_before_cursor = self.char_before_cursor
            current_line_before_cursor = self.current_line_before_cursor.strip()
            args = current_line_before_cursor.split(' ')
            args_len = len(args)
            if char_before_cursor == ' ':
                return args[-1]

            return args[-1] if args_len == 1 else args[-2]
        res = _()
        return res

    @property
    def word_before_cursor(self):
        """光标前的单词"""
        def _():
            if self.char_before_cursor == ' ':
                return ''
            text = self.current_line_before_cursor
            text = text.split(' ')[-1]
            if text in ('/', '~/'):
                return text
            return text.split('/')[-1]

        res = _()
        return res

    def get_completions(self, document, complete_event):
        self.document = document
        self.complete_event = complete_event
        for k in dir(document):
            if k.startswith('_'):
                continue
            val = getattr(document, k)
            if k.startswith('get'):
                continue
            if k.startswith('find'):
                continue
            if isinstance(val, int) or isinstance(val, str):
                setattr(self, k, val)

    def get_start_position(self):
        """获取文字补全的位置"""
        word = self.word_before_cursor
        if word in ('/', '~/'):
            return 0
        return -len(word)

    def filter(self, name):
        """过滤文件名"""
        name = name.lower()
        word = self.word_before_cursor.lower()
        if word in ('/', '~/'):
            return True
        if name.startswith(word):
            return True
        return False

    def search(self, words):
        """搜索"""
        keyword = self.word_before_cursor.lower()
        if keyword in ('/', '~/'):
            return words
        if not words:
            return words
        first_word = words[0]
        words_dict = {}
        if isinstance(first_word, Completion):
            words_dict = { o.text: o for o in words }
            words = [o.text for o in words]

        words = search(words, keyword)
        if isinstance(first_word, Completion):
            for i in range(len(words)):
                words[i] = words_dict[words[i]]

        return words




# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['wush',
 'wush.argument',
 'wush.cli',
 'wush.common',
 'wush.completion',
 'wush.config',
 'wush.model',
 'wush.web']

package_data = \
{'': ['*']}

install_requires = \
['Flask>=2.0.1,<3.0.0',
 'PyYAML>=6.0,<7.0',
 'browsercookie>=0.7.7,<0.8.0',
 'csarg>=0.1.0,<0.2.0',
 'loguru>=0.6.0,<0.7.0',
 'prompt-toolkit>=3.0.29,<4.0.0',
 'pydantic>=1.9.0,<2.0.0',
 'requests>=2.27.1,<3.0.0',
 'rich>=10.6.0,<11.0.0',
 'w3lib>=1.22.0,<2.0.0',
 'wpy>=0.6.1,<0.7.0']

entry_points = \
{'console_scripts': ['wush = wush.cli.main:main']}

setup_kwargs = {
    'name': 'wush',
    'version': '0.2.0',
    'description': 'Useful api client on terminal',
    'long_description': None,
    'author': 'wxnacy',
    'author_email': 'wxnacy@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6.2,<4.0.0',
}


setup(**setup_kwargs)



#!/usr/bin/env python
# -*- coding:utf-8 -*-
# Author: wxnacy@gmail.com
"""
命令补全
"""

import os
from prompt_toolkit.completion import Completion

from .base import BaseCompleter

__all__ = ['ExecutableCompleter']

class ExecutableCompleter(BaseCompleter):

    def get_completions(self, document, complete_event):
        super().get_completions(document, complete_event)
        last_word = document.text.split(' ')[-1]

        path = os.path.expanduser(last_word)
        dirname = os.path.dirname(path)

        for name in self.search(os.listdir(dirname)):
            start_position = self.get_start_position()
            yield Completion(name, start_position=start_position)




#!/usr/bin/env python
# -*- coding:utf-8 -*-
# Author: wxnacy(wxnacy@gmail.com)
"""
基础模块
"""
import json

from wpy.base import BaseObject
from wush.model.datatype import DataType
from wush.model.datatype import Object

__FILTER_SET_FIELDS__ = set(('AUTO_FORMAT', 'DEFAULT_DATATYPE',
    '__is_format__',  '__datatype_dict__'))

class Model(BaseObject):

    AUTO_FORMAT = False     # 是否自动 format
    # 字段默认的 datatype
    DEFAULT_DATATYPE = None

    # 装载 Model 的 datatype 类型字段
    __datatype_dict__ = dict()
    # 对象是否已经进行过 format 操作
    __is_format__ = False


    def __init__(self, **kwargs):
        # 初始化 __datatype_dict__
        self.__datatype_dict__ = {}
        self.__add_datatype_dict__(**self.get_cls_datatype_dict())
        super().__init__(**kwargs)

        # 将未赋值的字段设置为 None
        for key in self.__datatype_dict__.keys():
            if key not in kwargs:
            #  if hasattr(self, key) and isinstance(getattr(self, key), DataType):
                self.__dict__[key] = None

        # 判断自动 format
        if self.AUTO_FORMAT:
            self.format()

    def __setattr__(self, key, value):
        #  """重载设置值的方法"""
        # 如果字段的默认模型存在，则设置值之前先添加 datatype
        if self.DEFAULT_DATATYPE:
            default_datatype = self.__build_default_datatype()
            self.__add_datatype_dict__(**{key: default_datatype})

        self.__dict__[key] = value

    def get_cls_datatype_dict(self):
        item = {}
        for key, value in self.__class__.__dict__.items():
            if not isinstance(value, DataType):
                continue
            item[key] = value
        return item

    def __add_datatype_dict__(self, **kwargs):
        for k, v in kwargs.items():
            if k in __FILTER_SET_FIELDS__:
                continue
            if not isinstance(v, DataType):
                continue
            self.__datatype_dict__[k] = v

    def __build_default_datatype(self):
        """构建默认的 datatype"""
        return self.DEFAULT_DATATYPE

    def format(self):
        """将 datatype 字段进行格式化处理"""
        self._format(self)

    def _format(self, model):
        """嵌套 format"""
        # 判断是否有默认字段
        # 并对对象中存在的赋值添加到 __datatype_fields__ 中

        for k in model.__datatype_dict__.keys():
            # 格式化单个字段
            self._format_field(model, k)

        model.__dict__['__is_format__'] = True

    def _format_field(self, model, field):
        """格式化 field"""
        # 如果 field 不是 datatype 类型，直接返回
        #  datatype_fields = model.__get_datatype_fields()
        datatype_fields = model.__datatype_dict__
        dt_ins = datatype_fields.get(field)
        if not dt_ins:
            return

        try:
            value = model.__dict__.get(field)
            dt_ins.set_value(value)
        except:
            pass
        # 校验
        dt_ins.valid()
        set_val = dt_ins.value()
        # 清空数据
        dt_ins.clear()
        model.__dict__[field] = set_val
        # 嵌套 format
        if isinstance(set_val, Model):
            #  self._format(set_val)
            set_val.format()

    def dict(self):
        return self.to_dict()

    def to_dict(self):
        """
        将实例转为 dict 数据
        执行前需要保证实例已经执行过 format 方法或设置成员变量 AUTO_FORMAT=True
        """
        return self._to_dict(self)

    @classmethod
    def _to_dict(cls, model):
        """对 Model 进行循环 to_dict 操作"""
        data = {}
        #  for key in model.__get_datatype_fields().keys():
        for key in model.__datatype_dict__.keys():
            if not hasattr(model, key):
                continue
            origin_value = getattr(model, key)
            if isinstance(origin_value, Model):
                data[key] = origin_value.to_dict()
            else:
                data[key] = origin_value
        return data

    def to_json(self):
        return json.dumps(self, default=lambda o: o.to_dict(), sort_keys=True)

    def json(self):
        return self.to_json()

from wush.model import datatype
class Person(Model):
    AUTO_FORMAT = True
    id = datatype.Int(default=10)

class Book(Model):
    name = datatype.Str(default='book')

class User(Model):
    AUTO_FORMAT = True
    DEFAULT_DATATYPE = Object(model = Book)
    name = datatype.Str(default='wxnacy')

class Field(Model):
    doc = datatype.Str()
    _type = datatype.Object()
    value = datatype.Object()

class Json(Model):
    DEFAULT_DATATYPE = datatype.Object(model = Field)

    cust_field = datatype.Str(default='cust')

class Request(Model):
    json = datatype.Object(model=Json)

if __name__ == "__main__":
    #  p = Person()
    #  print(p.__datatype_dict__)
    #  print(p.id)
    #  assert p.id == 10

    #  b = Book()
    #  b.format()
    #  print(b.__datatype_dict__)
    #  print(b.name)
    #  assert b.name == 'book'

    u = User()
    assert u.name == 'wxnacy'
    u.book = { "name": "b1" }
    u.format()
    print(u.to_dict())
    #  assert u.book.name == 'b1'

    u1 = User()
    u1.book1 = { "name": "b2" }
    print(u1.to_dict())



#!/usr/bin/env python
# -*- coding:utf-8 -*-
# Author: wxnacy@gmail.com
"""

"""
from wush.common.loggers import create_logger
from wush.cli.command import Command

logger = create_logger('main')


def main():
    import time
    begin = time.time()
    cmd = Command()
    cmd.run()
    logger.info('wush time used: {}'.format(time.time() - begin))

if __name__ == "__main__":
    main()

#  index_model.__subclasses__():



#!/usr/bin/env python
# -*- coding:utf-8 -*-
"""

"""

import json

from wpy.base import BaseEnum
from wpy.base import BaseObject

from wush.web.enums import HeaderEnum
from wush.web.enums import ContentTypeEnum

class ResponseField(BaseEnum):
    HEADERS = 'headers'
    COOKIES = 'cookies'
    OK = 'ok'
    URL = 'url'
    CONTENT = 'content'
    TEXT = 'text'
    STATUS_CODE = 'status_code'


class ResponseClient(BaseObject):
    """返回客户端"""
    request_builder = None
    content = None
    text = None
    status_code = None
    url = None
    ok = None
    headers = None
    cookies = None

    def __init__(self, request_builder, response):
        self.request_builder = request_builder
        self.response = response
        for key in ResponseField.values():
            setattr(self, key, getattr(self.response, key))

    @property
    def content_type(self):
        return self.headers[HeaderEnum.CONTENT_TYPE.value]

    @property
    def location(self):
        return self.headers[HeaderEnum.LOCATION.value]

    @property
    def is_html(self):
        return ContentTypeEnum.TEXT_HTML.value in self.content_type

    @property
    def is_json(self):
        """判断结果是否为 json 格式"""
        if ContentTypeEnum.APPLICATION_JSON.value in self.content_type:
            return True
        return False

    def json(self, **kwargs):
        return self.response.json(**kwargs)

    def save(self, savepath=None):
        """保存"""
        if self.is_json:
            with open(savepath, 'w') as f:
                f.write(json.dumps(self.json(), indent=4))
        else:
            with open(savepath, 'wb') as f:
                f.write(self.content)

    def print(self):
        if self.is_json:
            print(json.dumps(self.json(), indent=4))
        else:
            print(self.text)


class ResponseHandler(BaseObject):

    _factory = {}

    @classmethod
    def register(cls, module, name):
        def decorate(func):
            key = f"{module}-{name}"
            cls._factory[key] = func
            return func
        return decorate

    @classmethod
    def get_handler(cls, module, name):
        """获取处理结果的方法

        """
        key = f"{module}-{name}"
        return cls._factory.get(key)



#!/usr/bin/env python
# -*- coding:utf-8 -*-
"""
请求历史
"""

import os
from wpy.base import BaseObject
from wpy.path import (
    read_dict, write_dict
)

from wush.config import load_config
from wush.web.response import ResponseClient

__all__ = ['History']

class HistoryModel(BaseObject):
    db = 'wush'
    table = 'history'

    version = ''
    request = dict()
    response = dict()

    def save(self, dirname):
        """保存历史"""
        if not os.path.exists(dirname):
            os.mkdir(dirname)
        filepath = os.path.join(dirname, self.version)
        write_dict(filepath, self.to_dict())

    @classmethod
    def find_one(cls, version, dirname):
        if not os.path.exists(dirname):
            os.mkdir(dirname)
        filepath = os.path.join(dirname, version)
        return read_dict(filepath)


class History(object):

    def __init__(self, *args, **kwargs):
        self.config = load_config()

    def save(self, response_client: ResponseClient):
        h = HistoryModel()
        h.version = response_client.request_builder.version
        h._id = h.version
        h.request = response_client.request_builder.to_requests()
        res = {
            'is_json': response_client.is_json,
            'text': response_client.text
        }
        if response_client.is_json:
            res['json'] = response_client.json()
        h.response = res
        h.save(self.config.api_history_dir)

    def read(self, version):
        """读取版本信息"""
        return HistoryModel.find_one(version, self.config.api_history_dir)




#!/usr/bin/env python
# -*- coding:utf-8 -*-
# Author: wxnacy@gmail.com
"""

"""
#  import argparse
import traceback
import multiprocessing as mp
import pygments

from csarg import CommandArgumentParser
from csarg import CommandArgumentParserFactory
from pygments.lexers.python import PythonLexer
from prompt_toolkit import print_formatted_text
from prompt_toolkit import PromptSession
from prompt_toolkit.auto_suggest import AutoSuggestFromHistory
from prompt_toolkit.formatted_text import PygmentsTokens
from prompt_toolkit.history import FileHistory
from wpy.functools import clock

from wush.common.loggers import create_logger
from wush.common.run_mode import RUN_MODE
from wush.common.constants import Constants
from wush.completion.command import CommandCompleter
from wush.config import load_config

from .exceptions import ContinueException
from .exceptions import CommnadNotFoundException
from .server import run_server


class Shell():
    logger = create_logger('Shell')

    parser_dict = {}
    parser = None
    _prompt_default = ''
    web_port = None
    session = None

    def __init__(self):
        self.parser = self._get_parser()
        self.config = load_config()
        self.session = PromptSession(
            completer=CommandCompleter(self.parser),
            # 设置历史记录文件
            history = FileHistory(Constants.HISTORY_PATH),
            auto_suggest = AutoSuggestFromHistory(),
            complete_in_thread=True
        )

    def _get_parser(self, cmd=None):
        if cmd not in self.parser_dict:
            parser = CommandArgumentParserFactory.build_parser(cmd)
            if isinstance(parser, CommandArgumentParser):
                parser.set_prompt(self.session)
            self.parser_dict[cmd] = parser
        return self.parser_dict[cmd]

    def run(self):
        RUN_MODE.set_shell()
        p = mp.Process(target=run_server, daemon=True)
        p.start()
        self._run_shell()
        p.terminate()

    def _run_shell(self):
        while True:
            try:
                #  left_prompt = 'wush/{space}/{module}> '.format(
                    #  space = self.config.space_name,
                    #  module = self.config.module_name
                #  )
                left_prompt = 'wush> '
                right_prompt = ''
                text = self.session.prompt(
                    left_prompt,
                    default = self._prompt_default,
                    rprompt = right_prompt,
                )
                self._run_once_time(text)
            except ContinueException:
                continue
            except CommnadNotFoundException:
                print('command not found: {}'.format(text))
            except KeyboardInterrupt:
                continue
            except EOFError:
                break
            except Exception as e:
                self._print('ERROR: ' + str(e))
                self.logger.error(traceback.format_exc())
            self._end_run()

        print('GoodBye!')

    def _end_run(self):
        self._prompt_default = ''

    @clock(fmt = Constants.CLOCK_FMT, logger_func = logger.info)
    def _run_once_time(self, text):
        """运行单次命令"""
        if not text:
            return
        parser = self._get_parser()
        args = parser.parse_args(text)
        cmd = args.cmd
        self.parser = self._get_parser(cmd)
        self.logger.info('run argparser %s', self.parser)

        self._run_base_cmd(text)

        self.logger.info(self.parser)
        if isinstance(self.parser, CommandArgumentParser):
            self.parser.run(text)
            return

        if not hasattr(self, '_' + cmd):
            raise CommnadNotFoundException()

        func = getattr(self, '_' + cmd)
        func(text)

    def _run_base_cmd(self, text):
        """运行基础命令"""
        if text.startswith('!'):
            text = text[1:]
            try:
                history_num = int(text)
                self.logger.info(history_num)
                cmd = self.get_history_by_num(history_num)
                self._prompt_default = cmd
            except:
                self.logger.error(traceback.format_exc())
                raise CommnadNotFoundException()
            else:
                raise ContinueException()

    def _exit(self, text):
        raise EOFError()

    def get_history_by_num(self, num):
        """获取历史命令"""
        items = self.session.history.get_strings()
        if len(items) < num:
            return None
        return items[num - 1]

    def _test(self, text):
        #  for k, v in os.environ.items():
            #  print(k, v)
        pass

    def _print(self, text):
        tokens = list(pygments.lex(text, lexer=PythonLexer()))
        print_formatted_text(PygmentsTokens(tokens), end='')




#!/usr/bin/env python
# -*- coding:utf-8 -*-
# Author: wxnacy@gmail.com
"""

"""
import argparse
import os
from collections import deque

from wpy.path import (
    read_dict, write_yml
)
from csarg import Action
from csarg import CommandArgumentParserFactory

from wush.config import load_config
from wush.argument.command import CmdArgumentParser

def init_argparse():
    """初始化参数"""
    parser = argparse.ArgumentParser(description='Wush command',)
    parser.add_argument('cmd', help='You can use run, body, env, module')
    parser.add_argument('-a', '--add', help='Config dir name', action='append')
    parser.add_argument('-v', '--verbose', help='Config dir name')
    parser.add_argument('-s', '--save', help='Config dir name')
    parser.add_argument('-f', '--file', action='store_true',
            help='Config dir name')
    return parser

@CommandArgumentParserFactory.register()
class EnvArgumentParser(CmdArgumentParser):
    cmd = 'env'

    def get_completions_after_argument(self, word_for_completion):
        """
        获取补全的单词列表
        :param word_for_completion: 补全需要的单词
        """
        if word_for_completion == '--add':
            words = []
            for k, v in self.config.env.items():
                text = '{}='.format(k)
                display = '{}={}'.format(k, v)
                words.append(dict(text = text, display = display))
            return words

        return super().get_completions_after_argument(word_for_completion)

    def _make_args_dict(self, args):
        res = super()._make_args_dict(args)
        args = list(filter(lambda x: x != '--save', args))
        if len(args) < 3:
            return res
        # 将所有成对的参数都放入参数列表
        d_args = deque(args[1:])
        d_args.rotate(-1)
        for i, (k, v) in enumerate(zip(args[1:], d_args)):
            if i % 2  == 1 or not k.startswith('--'):
                continue
            k = k.replace('--', '')
            res[k] = v
        return res

    @classmethod
    def default(cls):
        """
        初始化一个实例
        """
        item = cls()
        item.add_argument('cmd')
        item.add_argument('--save', action=Action.STORE_TRUE.value,
            help="保存环境变量")
        item.add_argument('--add', action=Action.APPEND.value,
            help="添加环境变量")

        item.sys_parser = init_argparse()
        item.config = load_config()

        return item

    def run(self, args):
        """执行"""
        self.logger.info(args)
        # TODO 替换本身的 argparse
        args_list = args.split(' ')
        args_list.append('--file')
        sys_args = self.sys_parser.parse_args(args_list)

        args = self.parse_args(args)
        for arg in self.get_arguments():
            self.logger.info(f'{self.cmd} {arg.name} {getattr(args, arg.name)}')
        if sys_args.add:
            for key_val in sys_args.add:
                key, val = key_val.split('=')
                self.config.add_env(key, val)
                # 写入当前环境变量
                # TODO 需要禁止修改敏感字段
                os.environ[key] = val

        output = {
            'headers': [
                { "display": "Key" },
                { "display": "Value" },
                ],
            "items": []
            }
        for k, v in self.config.env.items():
            line = (k, v)
            output['items'].append(line)
        self.config.function.print_table(output)

        # TODO 保存数据去掉 wapi
        # 保存数据
        if args.save:
            env_path = self.wapi.config.get_env_path()
            env_data = {}
            try:
                env_data = read_dict(env_path)
            except:
                pass
            for k, v in self.wapi.config.env.dict().items():
                if k in ('body_path',):
                    continue
                if k.isupper():
                    continue
                if not v:
                    continue
                if env_data.get(k) == v:
                    continue
                self.logger.info('Env save %s=%s', k, v)
                env_data[k] = v
            write_yml(env_path, env_data)

    def run_shell(self, args):
        self.run(args)

    def run_command(self, args):
        self.run(args)



#!/usr/bin/env python
# -*- coding:utf-8 -*-
# Author: wxnacy@gmail.com
"""
run 命令的参数解析
"""

from csarg import CommandArgumentParserFactory

from wush.cli.shell import Shell
from .command import CmdArgumentParser

__all__ = ['ShellArgumentParser']

@CommandArgumentParserFactory.register()
class ShellArgumentParser(CmdArgumentParser):
    cmd = 'shell'

    @classmethod
    def default(cls):
        """
        初始化一个实例
        """
        item = cls()
        item.add_argument('cmd')
        item.add_argument('-c', '--config', help='Config dir name')
        return item

    def run(self, text):
        shell = Shell()
        shell.run()
